from collections.abc import Iterator
from concurrent.futures import Future
from typing import TYPE_CHECKING, Annotated, Literal, Optional, Union, overload

from pydantic import Field, StrictStr

from snowflake.core import FQN, PollingOperation
from snowflake.core._common import (
    Clone,
    CreateMode,
    PointOfTime,
    SchemaObjectCollectionParent,
    SchemaObjectReferenceMixin,
)
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._operation import PollingOperations

from .._internal.telemetry import api_telemetry
from .._utils import tag_assignment_to_tag_tuple, tag_resource_to_tag_reference, tag_tuple_to_tag_assignment
from ..tag import TagResource, TagValue
from ._generated import SuccessResponse, TagAssignment, TagReference
from ._generated.api import AlertApi
from ._generated.models.alert import Alert
from ._generated.models.alert_clone import AlertClone
from ._generated.models.point_of_time import PointOfTime as AlertPointOfTime


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class AlertCollection(SchemaObjectCollectionParent["AlertResource"]):
    """Represents the collection operations on the Snowflake Alert resource.

    With this collection, you can create, update, iterate through, and fetch alerts that you have access to in the
    current context.
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, AlertResource)
        self._api = AlertApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @overload
    @api_telemetry
    def create(
        self, alert: str, *, clone_alert: Union[str, Clone], mode: CreateMode = CreateMode.error_if_exists
    ) -> "AlertResource": ...
    @overload
    @api_telemetry
    def create(
        self, alert: Alert, *, clone_alert: None, mode: CreateMode = CreateMode.error_if_exists
    ) -> "AlertResource": ...
    @api_telemetry
    def create(
        self,
        alert: Union[Alert, str],
        *,
        clone_alert: Optional[Union[str, Clone]] = None,
        mode: CreateMode = CreateMode.error_if_exists,
    ) -> "AlertResource":
        """Create an alert in Snowflake.

        There are two ways to create an alert: by cloning or by building from scratch.

        **Cloning an existing alert**

        Parameters
        __________
        alert: str
            The new alert's name
        clone_alert: str or Clone object
            The name of alert to be cloned, or a ``Clone`` object which would contain the name of the alert
            with support to clone at a specific time.
        mode: CreateMode, optional
            One of the following enum values:

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the alert already exists in Snowflake.  Equivalent to SQL ``create alert <name> ...``.

            ``CreateMode.or_replace``: Replace if the alert already exists in Snowflake. Equivalent to SQL
            ``create or replace alert <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the alert already exists in Snowflake.
            Equivalent to SQL ``create alert <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.

        Examples
        ________
        Cloning an alert instance:

        >>> alerts = schema.alerts
        >>> alerts.create(new_alert_name, clone_alert=alert_name_to_be_cloned, mode=CreateMode.if_not_exists)

        **Creating an alert from scratch**

        Parameters
        __________
        alert: Alert
            The details of ``Alert`` object, together with ``Alert``'s properties:
            name, schedule, condition, action ;
            comment, warehouse are optional.

        mode: CreateMode, optional
            One of the following enum values:

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the alert already exists in Snowflake.  Equivalent to SQL ``create alert <name> ...``.

            ``CreateMode.or_replace``: Replace if the alert already exists in Snowflake. Equivalent to SQL
            ``create or replace alert <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the alert already exists in Snowflake.
            Equivalent to SQL ``create alert <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.

        Examples
        ________
        Creating an alert instance:

        >>> alerts.create(
        ...     Alert(
        ...         name="my_alert",
        ...         warehouse="my_warehouse",
        ...         schedule="MinutesSchedule(minutes=1)",
        ...         condition="SELECT COUNT(*) FROM my_table > 100",
        ...         action="DROP TABLE my_table",
        ...     ),
        ...     mode=CreateMode.if_not_exists,
        ... )
        """
        self._create(alert=alert, clone_alert=clone_alert, mode=mode, async_req=False)
        return AlertResource(alert.name if isinstance(alert, Alert) else alert, self)

    @overload
    @api_telemetry
    def create_async(
        self, alert: str, *, clone_alert: Union[str, Clone], mode: CreateMode = CreateMode.error_if_exists
    ) -> PollingOperation["AlertResource"]: ...

    @overload
    @api_telemetry
    def create_async(
        self, alert: Alert, *, clone_alert: None, mode: CreateMode = CreateMode.error_if_exists
    ) -> PollingOperation["AlertResource"]: ...

    @api_telemetry
    def create_async(
        self,
        alert: Union[Alert, str],
        *,
        clone_alert: Optional[Union[str, Clone]] = None,
        mode: CreateMode = CreateMode.error_if_exists,
    ) -> PollingOperation["AlertResource"]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._create(alert=alert, clone_alert=clone_alert, mode=mode, async_req=True)
        return PollingOperation(
            future, lambda _: AlertResource(alert.name if isinstance(alert, Alert) else alert, self)
        )

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[StrictStr] = None,
        starts_with: Optional[StrictStr] = None,
        show_limit: Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]] = None,
        from_name: Optional[StrictStr] = None,
    ) -> Iterator[Alert]:
        """Iterate through ``Alert`` objects from Snowflake, filtering on any optional 'like' pattern.

        Parameters
        __________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).
        starts_with: str, optional
            String used to filter the command output based on the string of characters that appear
            at the beginning of the object name. Uses case-sensitive pattern matching.
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.
        from_name: str, optional
            Fetch rows only following the first row whose object name matches
            the specified string. This is case-sensitive and does not have to be the full name.

        Examples
        ________
        Showing all alerts that you have access to see:

        >>> alerts = alert_collection.iter()

        Showing information of the exact alert you want to see:

        >>> alerts = alert_collection.iter(like="your-alert-name")

        Showing alerts starting with 'your-alert-name-':

        >>> alerts = alert_collection.iter(like="your-alert-name-%")

        Using a for loop to retrieve information from iterator:

        >>> for alert in alerts:
        >>>     print(alert.name, alert.condition, alert.action)
        """
        alerts = self._api.list_alerts(
            database=self.database.name,
            var_schema=self.schema.name,
            like=like,
            starts_with=starts_with,
            show_limit=show_limit,
            from_name=from_name,
            async_req=False,
        )
        return iter(alerts)

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[StrictStr] = None,
        starts_with: Optional[StrictStr] = None,
        show_limit: Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]] = None,
        from_name: Optional[StrictStr] = None,
    ) -> PollingOperation[Iterator[Alert]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.list_alerts(
            database=self.database.name,
            var_schema=self.schema.name,
            like=like,
            starts_with=starts_with,
            show_limit=show_limit,
            from_name=from_name,
            async_req=True,
        )
        return PollingOperations.iterator(future)

    @overload
    def _create(
        self,
        alert: Union[Alert, str],
        clone_alert: Optional[Union[str, Clone]],
        mode: CreateMode,
        async_req: Literal[True],
    ) -> Future[SuccessResponse]: ...

    @overload
    def _create(
        self,
        alert: Union[Alert, str],
        clone_alert: Optional[Union[str, Clone]],
        mode: CreateMode,
        async_req: Literal[False],
    ) -> SuccessResponse: ...

    def _create(
        self, alert: Union[Alert, str], clone_alert: Optional[Union[str, Clone]], mode: CreateMode, async_req: bool
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        real_mode = CreateMode[mode].value

        if clone_alert:
            # create alert by clone
            if not isinstance(alert, str):
                raise TypeError("Alert has to be str for clone")

            pot: Optional[AlertPointOfTime] = None
            if isinstance(clone_alert, Clone) and isinstance(clone_alert.point_of_time, PointOfTime):
                pot = AlertPointOfTime.from_dict(clone_alert.point_of_time.to_dict())
            real_clone = Clone(source=clone_alert) if isinstance(clone_alert, str) else clone_alert
            req = AlertClone(point_of_time=pot, name=alert)

            source_alert_fqn = FQN.from_string(real_clone.source)
            return self._api.clone_alert(
                source_alert_fqn.database or self.database.name,
                source_alert_fqn.schema or self.schema.name,
                source_alert_fqn.name,
                alert_clone=req,
                create_mode=StrictStr(real_mode),
                async_req=async_req,
                target_database=self.database.name,
                target_schema=self.schema.name,
            )

        if not isinstance(alert, Alert):
            raise TypeError("alert has to be Alert object")
        return self._api.create_alert(
            self.database.name, self.schema.name, alert, create_mode=StrictStr(real_mode), async_req=async_req
        )


class AlertResource(SchemaObjectReferenceMixin[AlertCollection]):
    """Represents a reference to a Snowflake Alert resource.

    With this alert reference, you can fetch information about an alert, as well as perform certain
    actions on it.
    """

    def __init__(self, name: StrictStr, collection: AlertCollection) -> None:
        self.name = name
        self.collection = collection

    @api_telemetry
    def fetch(self) -> Alert:
        """Fetch the details of an alert.

        Examples
        ________
        Fetching an alert reference to print its name and query properties:

        >>> my_alert = alert_reference.fetch()
        >>> print(my_alert.name, my_alert.condition, my_alert.action)
        """
        return self.collection._api.fetch_alert(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def fetch_async(self) -> PollingOperation[Alert]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_alert(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperations.identity(future)

    @api_telemetry
    def drop(self, if_exists: bool = False) -> None:
        """Drop this alert.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this alert before drop.
            The default value is ``False``.

        Examples
        ________
        Deleting an alert using its reference, erroring if it doesn't exist:

        >>> alert_reference.drop()

        Deleting an alert using its reference if it exists:

        >>> alert_reference.drop(if_exists=True)
        """
        self.collection._api.delete_alert(
            self.database.name, self.schema.name, self.name, if_exists=if_exists, async_req=False
        )

    @api_telemetry
    def drop_async(self, if_exists: bool = False) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.delete_alert(
            self.database.name, self.schema.name, self.name, if_exists=if_exists, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def execute(self) -> None:
        """Execute an alert.

        Examples
        ________
        Use an alert reference to execute it:

        >>> alert_reference.execute()
        """
        self.collection._api.execute_alert(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def execute_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`execute`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.execute_alert(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def set_tags(self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None) -> None:
        """Set tags on an alert.

        Parameters
        __________
        tags: dict[TagResource, TagValue]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.set_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
        )

    @api_telemetry
    def set_tags_async(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`set_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.set_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def unset_tags(self, tag_resources: set[TagResource], if_exists: Optional[bool] = None) -> None:
        """Unset tags from an alert.

        Parameters
        __________
        tag_resources: set[TagResource]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.unset_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
        )

    @api_telemetry
    def unset_tags_async(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`unset_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.unset_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def get_tags(self, with_lineage: Optional[bool] = None) -> dict[TagResource, TagValue]:
        """Get the tag assignments for an alert.

        Returns all tags assigned to an alert. This operation requires an active warehouse.

        Parameters
        __________
        with_lineage: bool, optional
            Parameter that specifies whether tag assignments inherited by the object from its ancestors in securable
            object hierarchy should be returned as well: - `true`: All tags assigned to this object should be returned,
            inheritance included. - `false`: Only tags explicitly assigned to this object should be returned.
        """
        tag_assignments = self.collection._api.get_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            with_lineage=with_lineage,
        )
        return dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)

    @api_telemetry
    def get_tags_async(self, with_lineage: Optional[bool] = None) -> PollingOperation[dict[TagResource, TagValue]]:
        """An asynchronous version of :func:`get_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_tags(
            database=self.database.name,
            var_schema=self.schema.name,
            name=self.name,
            with_lineage=with_lineage,
            async_req=True,
        )
        return PollingOperation(
            future, lambda tag_assignments: dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)
        )



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


from contextlib import suppress

import pytest

from snowflake.core.compute_pool import ComputePool
from snowflake.core.exceptions import NotFoundError
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


@pytest.mark.min_sf_ver("8.37.0")
def test_create_or_alter(compute_pools, instance_family):
    cp_name = random_string(5, "test_cp_create_or_alter_")
    test_cp = ComputePool(
        name=cp_name,
        instance_family=instance_family,
        min_nodes=1,
        max_nodes=1,
        auto_resume=False,
        comment="created by temp_cp",
    )
    cp_ref = None
    try:
        # Test create when the ComputePool does not exist.
        cp_ref = compute_pools[cp_name]
        cp_ref.create_or_alter(test_cp)
        # currently it is not possible to create ComputePool as initially suspended using CoA so we suspend it manually
        compute_pools[cp_name].suspend()
        cp_list = compute_pools.iter(like=cp_name)
        result = next(cp_list)
        assert result.name == test_cp.name.upper()
        assert result.instance_family == test_cp.instance_family
        assert result.min_nodes == test_cp.min_nodes
        assert result.max_nodes == test_cp.max_nodes
        assert result.comment == test_cp.comment
        assert result.auto_suspend_secs == 3600
        assert result.auto_resume is False
        assert result.state in ("SUSPENDED", "STOPPING")

        # Make sure that issuing an empty alter doesn't create a malformed SQL
        cp_ref.create_or_alter(test_cp)

        # Test introducing property which was not set before
        test_cp_new_1 = ComputePool(
            name=cp_name,
            instance_family=instance_family,
            min_nodes=1,
            max_nodes=1,
            auto_resume=False,
            comment="created by temp_cp",
            auto_suspend_secs=30,
        )
        cp_ref.create_or_alter(test_cp_new_1)
        cp_list = compute_pools.iter(like=cp_name)
        result = next(cp_list)
        assert result.name == test_cp_new_1.name.upper()
        assert result.instance_family == test_cp_new_1.instance_family
        assert result.min_nodes == test_cp_new_1.min_nodes
        assert result.max_nodes == test_cp_new_1.max_nodes
        assert result.comment == test_cp_new_1.comment
        assert result.auto_suspend_secs == test_cp_new_1.auto_suspend_secs
        assert result.auto_resume is False
        assert result.state in ("SUSPENDED", "STOPPING")

        # Test altering the property which we set before
        test_cp_new_2 = ComputePool(
            name=cp_name,
            instance_family=instance_family,
            min_nodes=1,
            max_nodes=1,
            auto_resume=False,
            comment="new comment",
            auto_suspend_secs=90,
        )
        cp_ref.create_or_alter(test_cp_new_2)
        cp_list = compute_pools.iter(like=cp_name)
        result = next(cp_list)
        assert result.name == test_cp_new_2.name.upper()
        assert result.instance_family == test_cp_new_2.instance_family
        assert result.min_nodes == test_cp_new_2.min_nodes
        assert result.max_nodes == test_cp_new_2.max_nodes
        assert result.comment == test_cp_new_2.comment
        assert result.auto_suspend_secs == test_cp_new_2.auto_suspend_secs
        assert result.auto_resume is False
        assert result.state in ("SUSPENDED", "STOPPING")

        # Test not providing the property and checking that it is unset now and adding a new property at the same time
        test_cp_new_3 = ComputePool(
            name=cp_name,
            instance_family=instance_family,
            min_nodes=1,
            max_nodes=1,
            auto_resume=False,
            comment="comment",
        )
        cp_ref.create_or_alter(test_cp_new_3)
        cp_list = compute_pools.iter(like=cp_name)
        result = next(cp_list)
        assert result.name == test_cp_new_3.name.upper()
        assert result.instance_family == test_cp_new_3.instance_family
        assert result.min_nodes == test_cp_new_3.min_nodes
        assert result.max_nodes == test_cp_new_3.max_nodes
        assert result.comment == test_cp_new_3.comment
        assert result.auto_suspend_secs == 3600  # the default value
        assert result.auto_resume is False
        assert result.state in ("SUSPENDED", "STOPPING")

    finally:
        with suppress(NotFoundError):
            cp_ref.drop()



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt

from snowflake.core.cortex.analyst_service._generated.models.verified_query import VerifiedQuery, VerifiedQueryModel


class VerifiedQuerySuggestion(BaseModel):
    """A model object representing the VerifiedQuerySuggestion resource.

    Constructs an object of type VerifiedQuerySuggestion with the provided properties.

    Parameters
    __________
    score : float, optional
        Score of the suggestion, higher is better.
    vq_to_add : VerifiedQuery, optional

    vq_to_remove : VerifiedQuery, optional
    """

    score: Optional[Union[StrictFloat, StrictInt]] = None

    vq_to_add: Optional[VerifiedQuery] = None

    vq_to_remove: Optional[VerifiedQuery] = None

    __properties = ["score", "vq_to_add", "vq_to_remove"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> VerifiedQuerySuggestion:
        """Create an instance of VerifiedQuerySuggestion from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of vq_to_add
        if self.vq_to_add:
            _dict["vq_to_add"] = self.vq_to_add.to_dict()

        # override the default output from pydantic by calling `to_dict()` of vq_to_remove
        if self.vq_to_remove:
            _dict["vq_to_remove"] = self.vq_to_remove.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> VerifiedQuerySuggestion:
        """Create an instance of VerifiedQuerySuggestion from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return VerifiedQuerySuggestion.model_validate(obj)

        _obj = VerifiedQuerySuggestion.model_validate(
            {
                "score": obj.get("score"),
                "vq_to_add": VerifiedQuery.from_dict(obj.get("vq_to_add"))
                if obj.get("vq_to_add") is not None
                else None,
                "vq_to_remove": VerifiedQuery.from_dict(obj.get("vq_to_remove"))
                if obj.get("vq_to_remove") is not None
                else None,
            }
        )

        return _obj


class VerifiedQuerySuggestionModel:
    def __init__(
        self,  # optional properties
        score: Optional[float] = None,
        vq_to_add: Optional[VerifiedQuery] = None,
        vq_to_remove: Optional[VerifiedQuery] = None,
    ):
        """A model object representing the VerifiedQuerySuggestion resource.

        Constructs an object of type VerifiedQuerySuggestion with the provided properties.

        Parameters
        __________
        score : float, optional
            Score of the suggestion, higher is better.
        vq_to_add : VerifiedQuery, optional

        vq_to_remove : VerifiedQuery, optional
        """
        self.score = score
        self.vq_to_add = vq_to_add
        self.vq_to_remove = vq_to_remove

    __properties = ["score", "vq_to_add", "vq_to_remove"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return VerifiedQuerySuggestion(
            score=self.score,
            vq_to_add=self.vq_to_add._to_model() if self.vq_to_add is not None else None,
            vq_to_remove=self.vq_to_remove._to_model() if self.vq_to_remove is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> VerifiedQuerySuggestionModel:
        return VerifiedQuerySuggestionModel(
            score=model.score,
            vq_to_add=VerifiedQueryModel._from_model(model.vq_to_add) if model.vq_to_add else None,
            vq_to_remove=VerifiedQueryModel._from_model(model.vq_to_remove) if model.vq_to_remove else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a VerifiedQuerySuggestion.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the VerifiedQuerySuggestion object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> VerifiedQuerySuggestionModel:
        """Create an instance of VerifiedQuerySuggestion from a dict.

        This method constructs a VerifiedQuerySuggestion object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        VerifiedQuerySuggestion
            A VerifiedQuerySuggestion object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(VerifiedQuerySuggestion.from_dict(obj))


VerifiedQuerySuggestion._model_class = VerifiedQuerySuggestionModel



"""
Snowflake Task API.

The Snowflake Task API is a REST API that you can use to access, update, and perform certain actions on task resources in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr, field_validator


class TaskRun(BaseModel):
    """A model object representing the TaskRun resource.

    Constructs an object of type TaskRun with the provided properties.

    Parameters
    __________
    root_task_name : str
        The name of the root task in the current task run — **Read-only:** *any user-provided value will be ignored.*
    database_name : str
        The name of the current database for the task run — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str
        The name of the current schema for the task run — **Read-only:** *any user-provided value will be ignored.*
    state : str
        The current state of the task run — **Read-only:** *any user-provided value will be ignored.*
    scheduled_time : datetime
        The scheduled time for the task run — **Read-only:** *any user-provided value will be ignored.*
    next_scheduled_time : datetime
        The next upcoming time for the task run — **Read-only:** *any user-provided value will be ignored.*
    root_task_id : str
        The unique task ID for the root task — **Read-only:** *any user-provided value will be ignored.*
    graph_version : int
        The current version of the DAG on the task run — **Read-only:** *any user-provided value will be ignored.*
    run_id : int
        The unique ID for the current task run — **Read-only:** *any user-provided value will be ignored.*
    first_error_task_name : str, optional
        The name of the first task throwing an error in the task run — **Read-only:** *any user-provided value will be ignored.*
    first_error_code : int, optional
        The first error code thrown in the task run — **Read-only:** *any user-provided value will be ignored.*
    first_error_message : str, optional
        The first error message thrown in the task run — **Read-only:** *any user-provided value will be ignored.*
    query_start_time : datetime, optional
        The start time for the task run query — **Read-only:** *any user-provided value will be ignored.*
    completed_time : datetime, optional
        The time this task run was last completed — **Read-only:** *any user-provided value will be ignored.*
    """

    root_task_name: StrictStr

    database_name: StrictStr

    schema_name: StrictStr

    state: StrictStr

    first_error_task_name: Optional[StrictStr] = None

    first_error_code: Optional[StrictInt] = None

    first_error_message: Optional[StrictStr] = None

    scheduled_time: datetime

    query_start_time: Optional[datetime] = None

    next_scheduled_time: datetime

    completed_time: Optional[datetime] = None

    root_task_id: StrictStr

    graph_version: StrictInt

    run_id: StrictInt

    __properties = [
        "root_task_name",
        "database_name",
        "schema_name",
        "state",
        "first_error_task_name",
        "first_error_code",
        "first_error_message",
        "scheduled_time",
        "query_start_time",
        "next_scheduled_time",
        "completed_time",
        "root_task_id",
        "graph_version",
        "run_id",
    ]

    @field_validator("state")
    def state_validate_enum(cls, v):
        if v not in ("SCHEDULED", "EXECUTING", "SUCCEEDED", "FAILED", "CANCELLED", "SKIPPED"):
            raise ValueError(
                "must validate the enum values ('SCHEDULED','EXECUTING','SUCCEEDED','FAILED','CANCELLED','SKIPPED')"
            )
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> TaskRun:
        """Create an instance of TaskRun from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "root_task_name",
                    "database_name",
                    "schema_name",
                    "state",
                    "first_error_task_name",
                    "first_error_code",
                    "first_error_message",
                    "scheduled_time",
                    "query_start_time",
                    "next_scheduled_time",
                    "completed_time",
                    "root_task_id",
                    "graph_version",
                    "run_id",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> TaskRun:
        """Create an instance of TaskRun from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return TaskRun.model_validate(obj)

        _obj = TaskRun.model_validate(
            {
                "root_task_name": obj.get("root_task_name"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "state": obj.get("state"),
                "first_error_task_name": obj.get("first_error_task_name"),
                "first_error_code": obj.get("first_error_code"),
                "first_error_message": obj.get("first_error_message"),
                "scheduled_time": obj.get("scheduled_time"),
                "query_start_time": obj.get("query_start_time"),
                "next_scheduled_time": obj.get("next_scheduled_time"),
                "completed_time": obj.get("completed_time"),
                "root_task_id": obj.get("root_task_id"),
                "graph_version": obj.get("graph_version"),
                "run_id": obj.get("run_id"),
            }
        )

        return _obj


class TaskRunModel:
    def __init__(
        self,
        root_task_name: str,
        database_name: str,
        schema_name: str,
        state: str,
        scheduled_time: datetime,
        next_scheduled_time: datetime,
        root_task_id: str,
        graph_version: int,
        run_id: int,
        # optional properties
        first_error_task_name: Optional[str] = None,
        first_error_code: Optional[int] = None,
        first_error_message: Optional[str] = None,
        query_start_time: Optional[datetime] = None,
        completed_time: Optional[datetime] = None,
    ):
        """A model object representing the TaskRun resource.

        Constructs an object of type TaskRun with the provided properties.

        Parameters
        __________
        root_task_name : str
            The name of the root task in the current task run.
        database_name : str
            The name of the current database for the task run.
        schema_name : str
            The name of the current schema for the task run.
        state : str
            The current state of the task run.
        scheduled_time : datetime
            The scheduled time for the task run.
        next_scheduled_time : datetime
            The next upcoming time for the task run.
        root_task_id : str
            The unique task ID for the root task.
        graph_version : int
            The current version of the DAG on the task run.
        run_id : int
            The unique ID for the current task run.
        first_error_task_name : str, optional
            The name of the first task throwing an error in the task run.
        first_error_code : int, optional
            The first error code thrown in the task run.
        first_error_message : str, optional
            The first error message thrown in the task run.
        query_start_time : datetime, optional
            The start time for the task run query.
        completed_time : datetime, optional
            The time this task run was last completed.
        """
        self.root_task_name = root_task_name
        self.database_name = database_name
        self.schema_name = schema_name
        self.state = state
        self.first_error_task_name = first_error_task_name
        self.first_error_code = first_error_code
        self.first_error_message = first_error_message
        self.scheduled_time = scheduled_time
        self.query_start_time = query_start_time
        self.next_scheduled_time = next_scheduled_time
        self.completed_time = completed_time
        self.root_task_id = root_task_id
        self.graph_version = graph_version
        self.run_id = run_id

    __properties = [
        "root_task_name",
        "database_name",
        "schema_name",
        "state",
        "first_error_task_name",
        "first_error_code",
        "first_error_message",
        "scheduled_time",
        "query_start_time",
        "next_scheduled_time",
        "completed_time",
        "root_task_id",
        "graph_version",
        "run_id",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return TaskRun(
            root_task_name=self.root_task_name,
            database_name=self.database_name,
            schema_name=self.schema_name,
            state=self.state,
            first_error_task_name=self.first_error_task_name,
            first_error_code=self.first_error_code,
            first_error_message=self.first_error_message,
            scheduled_time=self.scheduled_time,
            query_start_time=self.query_start_time,
            next_scheduled_time=self.next_scheduled_time,
            completed_time=self.completed_time,
            root_task_id=self.root_task_id,
            graph_version=self.graph_version,
            run_id=self.run_id,
        )

    @classmethod
    def _from_model(cls, model) -> TaskRunModel:
        return TaskRunModel(
            root_task_name=model.root_task_name,
            database_name=model.database_name,
            schema_name=model.schema_name,
            state=model.state,
            first_error_task_name=model.first_error_task_name,
            first_error_code=model.first_error_code,
            first_error_message=model.first_error_message,
            scheduled_time=model.scheduled_time,
            query_start_time=model.query_start_time,
            next_scheduled_time=model.next_scheduled_time,
            completed_time=model.completed_time,
            root_task_id=model.root_task_id,
            graph_version=model.graph_version,
            run_id=model.run_id,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a TaskRun.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the TaskRun object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> TaskRunModel:
        """Create an instance of TaskRun from a dict.

        This method constructs a TaskRun object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        TaskRun
            A TaskRun object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(TaskRun.from_dict(obj))


TaskRun._model_class = TaskRunModel



from operator import attrgetter

import pytest


@pytest.mark.min_sf_ver("8.37.0")
def test_iter(notebooks, executable_notebook, temp_stage_case_sensitive):
    notebook_names = tuple(map(attrgetter("name"), notebooks.iter()))
    assert any(
        map(
            notebook_names.__contains__,
            (
                executable_notebook.name.upper(),  # for upper/lower case names
            ),
        )
    )


@pytest.mark.min_sf_ver("8.37.0")
def test_iter_like_with_from_name(notebooks, executable_notebook):
    notebook_names = tuple(map(attrgetter("name"), notebooks.iter(like="test_notebook%")))
    assert any(
        map(
            notebook_names.__contains__,
            (
                executable_notebook.name.upper(),  # for upper/lower case names
            ),
        )
    )
    notebook_def = executable_notebook.fetch()
    notebook_def.name += "_COPY"
    notebook_2 = notebooks.create(notebook_def)
    try:
        notebooks_from_name_limit = list(
            notebooks.iter(like=f"{notebook_def.name[:-5]}%", show_limit=1, from_name=notebook_def.name)
        )
        assert tuple(map(attrgetter("name"), notebooks_from_name_limit)) == (notebook_def.name,)
    finally:
        notebook_2.drop()



from snowflake.core.secret import (
    CloudProviderTokenSecret,
    GenericStringSecret,
    JwtKeyPairSecret,
    Oauth2Secret,
    PasswordSecret,
    SymmetricKeySecret,
)


test_cloud_provider_token_secret_template = CloudProviderTokenSecret(
    name="to_be_set",
    api_authentication="to_be_set",
    comment="Test cloud provider token secret",
)

test_generic_string_secret_template = GenericStringSecret(
    name="to_be_set",
    secret_string="test_secret_value_123",
    comment="Test generic string secret",
)

test_jwt_key_pair_secret_template = JwtKeyPairSecret(
    name="to_be_set",
    algorithm="RSA",
    key_length=4096,
    comment="Test JWT key pair secret",
)

test_oauth2_client_secret_template = Oauth2Secret(
    name="to_be_set",
    api_authentication="to_be_set",
    oauth_scopes=["read", "write"],
    comment="Test OAuth2 client flow secret",
)

test_oauth2_auth_code_secret_template = Oauth2Secret(
    name="to_be_set",
    api_authentication="to_be_set",
    oauth_refresh_token="test_refresh_token",
    oauth_refresh_token_expiry_time="2030-01-01 10:00:00",
    comment="Test OAuth2 auth code flow secret",
)

test_password_secret_template = PasswordSecret(
    name="to_be_set",
    username="snowman",
    password="test",
    comment="Test password secret",
)

test_symmetric_key_secret_template = SymmetricKeySecret(
    name="to_be_set",
    algorithm="GENERIC",
    comment="Test symmetric key secret",
)



from textwrap import dedent

import pytest

from snowflake.core.service import ServiceSpec, ServiceSpecInlineText, ServiceSpecStageFile
from snowflake.core.service._service import _parse_spec_path, _validate_inline_spec


_INLINE_SERVICE_SPEC = dedent("""
    spec:
      containers:
      - name: hello-world
        image: repo/hello-world:latest
""")
_INLINE_SERVICE_SPEC_NO_SPEC_FIELD = dedent("""
    containers:
    - name: hello-world
      image: repo/hello-world:latest
""")
_INLINE_SERVICE_SPEC_BAD_INDENTATION = dedent("""
    spec:
    containers:
        - name: hello-world
        image: repo/hello-world:latest
    - name: goodbye-world
    image: repo/goodbye-world:latest
""")


@pytest.mark.parametrize(
    "spec, expected",
    [
        ("@stage/spec.yaml", ServiceSpecStageFile(stage="stage", spec_file="spec.yaml")),
        ("@stage/path/to/spec.yaml", ServiceSpecStageFile(stage="stage", spec_file="path/to/spec.yaml")),
        (" @stage/path/with/spaces  ", ServiceSpecStageFile(stage="stage", spec_file="path/with/spaces")),
        (_INLINE_SERVICE_SPEC, ServiceSpecInlineText(spec_text=_INLINE_SERVICE_SPEC.rstrip())),
    ],
)
def test_service_spec(spec, expected):
    assert ServiceSpec(spec) == expected


@pytest.mark.parametrize(
    "spec",
    [
        "@stage",  # no forward slash followed by file path
        "@stage/",  # no file path after forward slash
        "@stage//spec.yaml",  # doubled forward slash
        "@stage/spec.yaml/",  # trailing forward slash
        "stage/path/spec.yaml",  # no leading @, gets interpreted as invalid inline spec
        _INLINE_SERVICE_SPEC_NO_SPEC_FIELD,
        _INLINE_SERVICE_SPEC_BAD_INDENTATION,
    ],
)
def test_service_spec_invalid(spec):
    with pytest.raises(ValueError):
        ServiceSpec(spec)


@pytest.mark.parametrize(
    "spec_path, stage, path",
    [("stage/spec.yaml", "stage", "spec.yaml"), ("stage/path/to/spec.yaml", "stage", "path/to/spec.yaml")],
)
def test_parse_spec_path(spec_path, stage, path):
    assert _parse_spec_path(spec_path) == (stage, path)


@pytest.mark.parametrize(
    "spec_path",
    [
        "stage",  # no forward slash followed by file path
        "stage/",  # no file path after forward slash
        "stage//invalid.yaml",  # doubled forward slash
        "stage/spec.yaml/",  # trailing forward slash
        "/path/to/spec.yaml",  # leading forward slash
        "stage/<invalid>/spec.yaml",  # invalid characters in file path
        "?name/spec.yaml"  # invalid character in stage name
        r"stage\path\to\spec.yaml",  # backslash instead of forward slash
        "",  # empty
    ],
)
def test_parse_spec_path_invalid(spec_path):
    with pytest.raises(ValueError):
        _parse_spec_path(spec_path)


def test_validate_inline_spec():
    assert _validate_inline_spec(_INLINE_SERVICE_SPEC)


@pytest.mark.parametrize(
    "spec_str",
    [
        _INLINE_SERVICE_SPEC_NO_SPEC_FIELD,
        _INLINE_SERVICE_SPEC_BAD_INDENTATION,
        "",  # empty
    ],
)
def test_validate_inline_spec_invalid(spec_str):
    assert not _validate_inline_spec(spec_str)



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.

"""Manages Snowflake Tasks.

Example:
    >>> tasks: TaskCollection = root.databases["mydb"].schemas["myschema"].tasks
    >>> mytask = tasks.create(Task("mytask", definition="select 1"))
    >>> task_iter = tasks.iter(like="my%")
    >>> mytask = tasks["mytask"]
    >>> # Then call other APIs to manage this task.
    >>> mytask.resume()
    >>> mytask.suspend()
    >>> an_existing_task = tasks["an_existing_task"]
    >>> an_existing_task.suspend()

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from ._generated.models.task_run import TaskRun
from ._task import Cron, StoredProcedureCall, Task, TaskCollection, TaskResource


__all__ = [
    "Cron",
    "StoredProcedureCall",
    "Task",
    "TaskCollection",
    "TaskResource",
    "TaskRun",
]



from unittest import mock

from snowflake.core import PollingOperation
from snowflake.core.schema import Schema, SchemaResource
from snowflake.core.schema._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


def test_create_schema(fake_root, schemas):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas?createMode=errorIfExists")
    kwargs = extra_params(
        query_params=[("createMode", "errorIfExists")],
        body={"name": "my_schema", "kind": "PERMANENT", "managed_access": False, "dropped_on": None},
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema_res = schemas.create(Schema(name="my_schema"))
        assert isinstance(schema_res, SchemaResource)
        assert schema_res.name == "my_schema"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schemas.create_async(Schema(name="my_schema"))
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_schema"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_schema(fake_root, schemas):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        schemas.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = schemas.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_create_or_alter_schema(fake_root, schema):
    args = (fake_root, "PUT", BASE_URL + "/databases/my_db/schemas/my_schema")
    kwargs = extra_params(body={"name": "my_schema", "kind": "PERMANENT", "managed_access": False, "dropped_on": None})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema.create_or_alter(Schema(name="my_schema"))
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schema.create_or_alter_async(Schema(name="my_schema"))
        assert isinstance(op, PollingOperation)
        schema_res = op.result()
        assert schema_res.name == "my_schema"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_schema(fake_root, schema):
    from snowflake.core.schema._generated.models import Schema

    model = Schema(name="my_schema")
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        schema.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = schema.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == Schema(name="my_schema").to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_schema(fake_root, schema):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schema.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_undrop_schema(fake_root, schema):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema:undrop")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema.undrop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schema.undrop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_set_tags(fake_root, schema, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schema.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, schema, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        schema.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = schema.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, schema):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert schema.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = schema.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



from unittest import mock

import pytest

from snowflake.core import CreateMode, PollingOperation
from snowflake.core.sequence import Sequence, SequenceResource

from ...utils import BASE_URL, extra_params, mock_http_response


@pytest.fixture
def sequences(schema):
    return schema.sequences


@pytest.fixture
def sequence(sequences) -> SequenceResource:
    return sequences["my_sequence"]


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
SEQUENCE = Sequence(name="my_sequence", start=1, increment=1, ordered=True, comment="Test sequence")


def test_create_sequence(fake_root, sequences):
    kwargs = extra_params(
        body={
            "name": "my_sequence",
            "start": 1,
            "increment": 1,
            "ordered": True,
            "comment": "Test sequence",
        },
    )
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/sequences")

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        s_res = sequences.create(SEQUENCE)
        assert isinstance(s_res, SequenceResource)
        assert s_res.name == "my_sequence"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = sequences.create_async(SEQUENCE)
        assert isinstance(op, PollingOperation)
        s_res = op.result()
        assert s_res.name == "my_sequence"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_sequence(fake_root, sequences):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/sequences")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        sequences.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = sequences.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_sequence(fake_root, sequence):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/sequences/my_sequence")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(SEQUENCE.to_json())
        s = sequence.fetch()
        assert s.to_dict() == SEQUENCE.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(SEQUENCE.to_json())
        op = sequence.fetch_async()
        assert isinstance(op, PollingOperation)
        s = op.result()
        assert s.to_dict() == SEQUENCE.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_sequence(fake_root, sequence):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/sequences/my_sequence")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        sequence.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = sequence.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_sequence(fake_root, sequence, sequences):
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/sequences/my_sequence:rename?targetName=new_sequence",
    )
    kwargs = extra_params(query_params=[("targetName", "new_sequence")])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        sequence.rename("new_sequence")
        assert sequence.name == "new_sequence"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        s_res = sequences["another_sequence"]
        op = s_res.rename_async("new_sequence")
        assert isinstance(op, PollingOperation)
        op.result()
        assert s_res.name == "new_sequence"
    args2 = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/sequences/another_sequence:rename?targetName=new_sequence",
    )
    mocked_request.assert_called_once_with(*args2, **kwargs)


def test_rename_sequence_with_options(fake_root, sequence):
    args = (
        fake_root,
        "POST",
        BASE_URL
        + "/databases/my_db/schemas/my_schema/sequences/my_sequence:rename?ifExists=True&targetDatabase=other_db&targetSchema=other_schema&targetName=new_sequence",
    )
    kwargs = extra_params(
        query_params=[
            ("ifExists", True),
            ("targetDatabase", "other_db"),
            ("targetSchema", "other_schema"),
            ("targetName", "new_sequence"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        sequence.rename(
            "new_sequence",
            target_database="other_db",
            target_schema="other_schema",
            if_exists=True,
        )
        assert sequence.name == "new_sequence"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_clone_sequence(fake_root, sequence):
    sequence_clone = Sequence(name="sequence_clone")
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/sequences/my_sequence:clone",
    )
    kwargs = extra_params(
        body={
            "name": "sequence_clone",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        sequence.clone(sequence_clone)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = sequence.clone_async(sequence_clone)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_clone_sequence_with_options(fake_root, sequence):
    args = (
        fake_root,
        "POST",
        BASE_URL
        + "/databases/my_db/schemas/my_schema/sequences/my_sequence:clone?createMode=orReplace&targetDatabase=other_db&targetSchema=other_schema",
    )
    kwargs = extra_params(
        query_params=[
            ("createMode", "orReplace"),
            ("targetDatabase", "other_db"),
            ("targetSchema", "other_schema"),
        ],
        body={
            "name": "sequence_clone",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        sequence.clone(
            Sequence(name="sequence_clone"),
            create_mode=CreateMode.or_replace,
            target_database="other_db",
            target_schema="other_schema",
        )
    mocked_request.assert_called_once_with(*args, **kwargs)



import pytest

from snowflake.core._internal.utils import normalize_and_unquote_name
from tests.integ.external_volume.utils import (
    assert_basic,
    get_aws_external_volume_template,
    get_aws_external_volume_template_case_sensitive,
    get_azure_external_volume_template,
    get_gcs_external_volume_template,
)


pytestmark = [pytest.mark.internal_only]


def test_fetch_external_volume(external_volumes, setup_credentials_fixture):
    try:
        external_volume_to_drop = []
        # happy path aws
        external_volume = get_aws_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]
        external_volume_object = external_volume_handler.fetch()
        assert_basic(external_volume_object, external_volume)
        assert external_volume_object.name == external_volume.name.upper()
        assert external_volume_object.comment == external_volume.comment
        assert len(external_volume_object.storage_locations) == 2
        assert external_volume_object.storage_locations[0].name == "abcd-my-s3-us-west-2"
        assert external_volume_object.storage_locations[0].storage_base_url == "s3://MY_EXAMPLE_BUCKET/"
        assert (
            external_volume_object.storage_locations[0].storage_aws_role_arn == "arn:aws:iam::123456789022:role/myrole"
        )
        assert external_volume_object.storage_locations[0].encryption.type == "AWS_SSE_KMS"
        assert (
            external_volume_object.storage_locations[0].encryption.kms_key_id == "1234abcd-12ab-34cd-56ef-1234567890ab"
        )
        assert external_volume_object.storage_locations[1].name == "abcd-my-s3-us-west-3"
        assert external_volume_object.storage_locations[1].storage_base_url == "s3://MY_EXAMPLE_BUCKET_1/"
        assert (
            external_volume_object.storage_locations[1].storage_aws_role_arn == "arn:aws:iam::123456789022:role/myrole"
        )
        assert external_volume_object.storage_locations[1].encryption.type == "AWS_SSE_KMS"
        assert (
            external_volume_object.storage_locations[1].encryption.kms_key_id == "1234abcd-12ab-34cd-56ef-1234567890ab"
        )

        # happy path aws case sensitive
        external_volume = get_aws_external_volume_template_case_sensitive()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]
        external_volume_object = external_volume_handler.fetch()
        assert external_volume_object.name == normalize_and_unquote_name(external_volume.name)
        assert external_volume_object.comment == external_volume.comment

        # happy path azure
        external_volume = get_azure_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]
        external_volume_object = external_volume_handler.fetch()
        assert external_volume_object.name == external_volume.name.upper()
        assert_basic(external_volume_object, external_volume)
        assert len(external_volume_object.storage_locations) == 1
        assert external_volume_object.storage_locations[0].name == "abcd-my-azure-northeurope"
        assert (
            external_volume_object.storage_locations[0].storage_base_url
            == "azure://exampleacct.blob.core.windows.net/my_container_northeurope/"
        )
        assert external_volume_object.storage_locations[0].azure_tenant_id == "a123b4c5-1234-123a-a12b-1a23b45678c9"

        # happy path gcs
        external_volume = get_gcs_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]
        external_volume_object = external_volume_handler.fetch()
        assert external_volume_object.name == external_volume.name.upper()
        assert_basic(external_volume_object, external_volume)
        assert len(external_volume_object.storage_locations) == 1
        assert external_volume_object.storage_locations[0].name == "abcd-my-us-east-1"
        assert external_volume_object.storage_locations[0].storage_base_url == "gcs://mybucket1/path1/"
        assert external_volume_object.storage_locations[0].encryption.type == "GCS_SSE_KMS"
        assert (
            external_volume_object.storage_locations[0].encryption.kms_key_id == "1234abcd-12ab-34cd-56ef-1234567890ab"
        )

    finally:
        for external_volume_handler in external_volume_to_drop:
            external_volume_handler.drop(if_exists=True)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from io import BytesIO
from textwrap import dedent

import pytest

from snowflake.core.service import (
    Service,
    ServiceContainer,
    ServiceEndpoint,
    ServiceInstance,
    ServiceRole,
    ServiceSpecStageFile,
)
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


@pytest.mark.usefixtures("qa_mode_enabled")
def test_get_commands(
    services, session, imagerepo, shared_compute_pool, database, schema, setup_with_connector_execution
):
    # Mock resource set status
    with setup_with_connector_execution(["alter session set ENABLE_MOCK_RESOURCE_SET_STATUS=true"], []):
        # Create a service
        service_name = random_string(5, "test_service_")
        stage_name = random_string(5, "test_stage_")
        session.sql(f"create temp stage {stage_name};").collect()
        spec_file = "spec.yaml"
        stage_file = f"@{stage_name}"
        spec = f"{stage_file}/{spec_file}"
        session.file.put_stream(
            BytesIO(
                dedent(f"""
                        spec:
                          containers:
                          - name: main-container
                            image: {imagerepo}/hello-world:latest
                          - name: other-container
                            image: {imagerepo}/hello-world:latest
                          endpoints:
                          - name: echo
                            port: 8080
                        serviceRoles:
                        - name: role1
                          endpoints:
                          - echo
                        """).encode()
            ),
            spec,
        )
        test_service = Service(
            name=service_name,
            compute_pool=shared_compute_pool,
            spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
            min_instances=1,
            max_instances=1,
        )
        svc = services.create(test_service)

        try:
            # Verify the output of get_containers command
            expected_containers = [
                ServiceContainer.from_dict(
                    {
                        "database_name": database.name.upper(),
                        "schema_name": schema.name.upper(),
                        "service_name": service_name.upper(),
                        "instance_id": "0",
                        "container_name": "main-container",
                        "status": "DONE",
                        "message": "DONE",
                        "image_name": f"{imagerepo}/hello-world:latest",
                        "image_digest": None,
                        "restart_count": 0,
                        "start_time": "2023-01-01T00:00:00Z",
                    }
                ),
                ServiceContainer.from_dict(
                    {
                        "database_name": database.name.upper(),
                        "schema_name": schema.name.upper(),
                        "service_name": service_name.upper(),
                        "instance_id": "0",
                        "container_name": "other-container",
                        "status": "DONE",
                        "message": "DONE",
                        "image_name": f"{imagerepo}/hello-world:latest",
                        "image_digest": None,
                        "restart_count": 0,
                        "start_time": "2023-01-01T00:00:00Z",
                    }
                ),
            ]

            actual_containers = list(svc.get_containers())
            assert actual_containers == expected_containers

            # Verify the output of get_instances command
            expected_instances = [
                ServiceInstance.from_dict(
                    {
                        "database_name": database.name.upper(),
                        "schema_name": schema.name.upper(),
                        "service_name": service_name.upper(),
                        "instance_id": "0",
                        "status": "SUCCEEDED",
                        "spec_digest": "",
                        "creation_time": "2023-01-01T00:00:00Z",
                        "start_time": "2023-01-01T00:00:00Z",
                    }
                )
            ]

            actual_instances = list(svc.get_instances())
            assert actual_instances == expected_instances

            # Verify the output of get_roles command
            expected_roles = [
                ServiceRole.from_dict(
                    {"created_on": "1967-06-23T07:00:00.123+00:00", "name": "ALL_ENDPOINTS_USAGE", "comment": None}
                ),
                ServiceRole.from_dict(
                    {"created_on": "1967-06-23T07:00:00.123+00:00", "name": "ROLE1", "comment": None}
                ),
            ]

            actual_roles = list(svc.get_roles())
            assert actual_roles == expected_roles

            # Verify the output of get_endpoints command
            expected_endpoints = [
                ServiceEndpoint.from_dict(
                    {
                        "name": "echo",
                        "port": 8080,
                        "port_range": None,
                        "protocol": "HTTP",
                        "is_public": False,
                        "ingress_url": None,
                    }
                )
            ]

            actual_endpoints = list(svc.get_endpoints())
            assert actual_endpoints == expected_endpoints
        finally:
            svc.drop()



from itertools import combinations
from unittest import mock
from unittest.mock import MagicMock

import pytest

from snowflake.core import Clone, CreateMode
from snowflake.core.iceberg_table import IcebergTable, IcebergTableCollection, IcebergTableResource
from snowflake.core.iceberg_table._generated import (
    ConvertToManagedIcebergTableRequest,
    IcebergTableAsSelect,
    IcebergTableClone,
    IcebergTableFromAWSGlueCatalog,
    IcebergTableFromDelta,
    IcebergTableFromIcebergFiles,
    IcebergTableFromIcebergRest,
    IcebergTableLike,
    RefreshIcebergTableRequest,
)
from snowflake.core.schema import SchemaResource


@pytest.fixture
def _mock_collection():
    return MagicMock(database=MagicMock())


@pytest.fixture
def _mock_schema(_mock_collection):
    return SchemaResource("test_schema", _mock_collection)


@pytest.fixture
def _mock_iceberg_tables_collection(_mock_schema):
    return IcebergTableCollection(schema=_mock_schema)


@pytest.fixture
def _mock_api():
    with mock.patch("snowflake.core.iceberg_table._iceberg_table.IcebergTableApi") as mock_api:
        yield mock_api.return_value


def parametrize_if_exists():
    return pytest.mark.parametrize("if_exists", [True, False])


def parametrize_mode():
    return pytest.mark.parametrize("mode", [CreateMode.error_if_exists, CreateMode.if_not_exists])


def parametrize_copy_grants():
    return pytest.mark.parametrize("copy_grants", [True, False])


def parametrize_async():
    return pytest.mark.parametrize("is_async", [True, False])


class TestIcebergTableResource:
    @parametrize_async()
    def test_fetch_iceberg_table(self, _mock_collection, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        if not is_async:
            table.fetch()
        else:
            table.fetch_async()
        _mock_collection._api.fetch_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name, _mock_collection.schema.name, "my_table", async_req=is_async
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_drop_iceberg_table(self, _mock_collection, if_exists, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        if not is_async:
            table.drop(if_exists=if_exists)
        else:
            table.drop_async(if_exists=if_exists)
        _mock_collection._api.drop_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name,
            _mock_collection.schema.name,
            "my_table",
            if_exists=if_exists,
            async_req=is_async,
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_convert_to_managed_iceberg_table(self, _mock_collection, if_exists, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        convert_fn = table.convert_to_managed_async if is_async else table.convert_to_managed
        convert_fn(base_location="test_base_location", storage_serialization_policy="COMPATIBLE", if_exists=if_exists)
        _mock_collection._api.convert_to_managed_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name,
            _mock_collection.schema.name,
            "my_table",
            convert_to_managed_iceberg_table_request=ConvertToManagedIcebergTableRequest(
                base_location="test_base_location", storage_serialization_policy="COMPATIBLE"
            ),
            if_exists=if_exists,
            async_req=is_async,
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_refresh_iceberg_table(self, _mock_collection, if_exists, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        refresh_fn = table.refresh_async if is_async else table.refresh
        refresh_fn(metadata_file_relative_path="test_metadata_file_relative_path", if_exists=if_exists)
        _mock_collection._api.refresh_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name,
            _mock_collection.schema.name,
            "my_table",
            refresh_iceberg_table_request=RefreshIcebergTableRequest(
                metadata_file_relative_path="test_metadata_file_relative_path"
            ),
            if_exists=if_exists,
            async_req=is_async,
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_resume_recluster_table(self, _mock_collection, if_exists, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        if not is_async:
            table.resume_recluster(if_exists=if_exists)
        else:
            table.resume_recluster_async(if_exists=if_exists)
        _mock_collection._api.resume_recluster_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name,
            _mock_collection.schema.name,
            "my_table",
            if_exists=if_exists,
            async_req=is_async,
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_suspend_recluster_table(self, _mock_collection, if_exists, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        if not is_async:
            table.suspend_recluster(if_exists=if_exists)
        else:
            table.suspend_recluster_async(if_exists=if_exists)
        _mock_collection._api.suspend_recluster_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name,
            _mock_collection.schema.name,
            "my_table",
            if_exists=if_exists,
            async_req=is_async,
        )

    @parametrize_async()
    def test_undrop_table(self, _mock_collection, is_async):
        table = IcebergTableResource(name="my_table", collection=_mock_collection)
        if not is_async:
            table.undrop()
        else:
            table.undrop_async()
        _mock_collection._api.undrop_iceberg_table.assert_called_once_with(
            _mock_collection.schema.database.name, _mock_collection.schema.name, "my_table", async_req=is_async
        )


class TestIcebergTableCollection:
    @parametrize_async()
    def test_iter(self, _mock_api, _mock_schema, _mock_iceberg_tables_collection, is_async):
        iter_fn = _mock_iceberg_tables_collection.iter_async if is_async else _mock_iceberg_tables_collection.iter
        iter_fn(like="%my_table", starts_with="bar", show_limit=42, from_name="foo")

        _mock_api.list_iceberg_tables.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            like="%my_table",
            starts_with="bar",
            show_limit=42,
            from_name="foo",
            async_req=is_async,
        )

    @pytest.mark.parametrize(
        "options",
        combinations(
            [
                "as_select",
                "like",
                "from_aws_glue_catalog",
                "from_delta",
                "from_iceberg_files",
                "from_iceberg_rest",
                "clone_iceberg_table",
            ],
            2,
        ),
    )
    def test_validate_iceberg_table(self, options, _mock_iceberg_tables_collection):
        kwargs = {o: True for o in options}
        with pytest.raises(ValueError) as err:
            _mock_iceberg_tables_collection.create(iceberg_table=IcebergTable(name="foo"), **kwargs)
        assert "are mutually exclusive" in str(err)

    @parametrize_mode()
    @parametrize_copy_grants()
    @parametrize_async()
    def test_create_iceberg_table_as_select(
        self, _mock_api, _mock_iceberg_tables_collection, mode, copy_grants, is_async
    ):
        iceberg_table = IcebergTable(name="fooObject", base_location="@stage/path")

        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, as_select="select * from bar", mode=mode, copy_grants=copy_grants)

        _mock_api.create_snowflake_managed_iceberg_table_as_select.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            query="select * from bar",
            iceberg_table_as_select=IcebergTableAsSelect(**iceberg_table.to_dict()),
            create_mode=mode,
            copy_grants=copy_grants,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_copy_grants()
    @parametrize_async()
    def test_create_iceberg_table_like(self, _mock_api, _mock_iceberg_tables_collection, mode, copy_grants, is_async):
        iceberg_table = IcebergTable(name="fooObject")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, like="barTable", mode=mode, copy_grants=copy_grants)

        _mock_api.create_snowflake_managed_iceberg_table_like.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            name="barTable",
            iceberg_table_like=IcebergTableLike(**iceberg_table.to_dict()),
            create_mode=mode,
            copy_grants=copy_grants,
            target_database=_mock_iceberg_tables_collection.schema.database.name,
            target_schema=_mock_iceberg_tables_collection.schema.name,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_async()
    def test_create_iceberg_table_from_aws_glue_catalog(
        self, _mock_api, _mock_iceberg_tables_collection, mode, is_async
    ):
        iceberg_table = IcebergTable(name="fooObject", catalog_table_name="fooCatalogTable")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, from_aws_glue_catalog=True, mode=mode)

        _mock_api.create_unmanaged_iceberg_table_from_aws_glue_catalog.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            iceberg_table_from_aws_glue_catalog=IcebergTableFromAWSGlueCatalog(**iceberg_table.to_dict()),
            create_mode=mode,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_async()
    def test_create_iceberg_table_from_delta(self, _mock_api, _mock_iceberg_tables_collection, mode, is_async):
        iceberg_table = IcebergTable(name="fooObject", base_location="@stage/path")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, from_delta=True, mode=mode)

        _mock_api.create_unmanaged_iceberg_table_from_delta.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            iceberg_table_from_delta=IcebergTableFromDelta(**iceberg_table.to_dict()),
            create_mode=mode,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_async()
    def test_create_iceberg_table_from_iceberg_files(self, _mock_api, _mock_iceberg_tables_collection, mode, is_async):
        iceberg_table = IcebergTable(name="fooObject", metadata_file_path="@stage/path")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, from_iceberg_files=True, mode=mode)

        _mock_api.create_unmanaged_iceberg_table_from_iceberg_files.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            iceberg_table_from_iceberg_files=IcebergTableFromIcebergFiles(**iceberg_table.to_dict()),
            create_mode=mode,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_async()
    def test_create_iceberg_table_from_iceberg_rest(self, _mock_api, _mock_iceberg_tables_collection, mode, is_async):
        iceberg_table = IcebergTable(name="fooObject", catalog_table_name="fooCatalogTable")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, from_iceberg_rest=True, mode=mode)

        _mock_api.create_unmanaged_iceberg_table_from_iceberg_rest.assert_called_once_with(
            database=_mock_iceberg_tables_collection.schema.database.name,
            var_schema=_mock_iceberg_tables_collection.schema.name,
            iceberg_table_from_iceberg_rest=IcebergTableFromIcebergRest(**iceberg_table.to_dict()),
            create_mode=mode,
            async_req=is_async,
        )

    @parametrize_mode()
    @parametrize_copy_grants()
    @parametrize_async()
    def test_create_iceberg_table_clone_iceberg_table(
        self, _mock_api, _mock_iceberg_tables_collection, mode, copy_grants, is_async
    ):
        iceberg_table = IcebergTable(name="fooObject")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(
            iceberg_table=iceberg_table,
            clone_iceberg_table=Clone(source="barTable"),
            mode=mode,
            copy_grants=copy_grants,
        )

        _mock_api.clone_snowflake_managed_iceberg_table.assert_called_once_with(
            _mock_iceberg_tables_collection.schema.database.name,
            _mock_iceberg_tables_collection.schema.name,
            "barTable",
            iceberg_table_clone=IcebergTableClone(point_of_time=None, name=iceberg_table.name),
            create_mode=mode,
            async_req=is_async,
            target_database=_mock_iceberg_tables_collection.schema.database.name,
            target_schema=_mock_iceberg_tables_collection.schema.name,
            copy_grants=copy_grants,
        )

    @parametrize_mode()
    @parametrize_copy_grants()
    @parametrize_async()
    def test_create_iceberg_table(self, _mock_api, _mock_iceberg_tables_collection, mode, copy_grants, is_async):
        iceberg_table = IcebergTable(name="fooObject")
        create_fn = _mock_iceberg_tables_collection.create_async if is_async else _mock_iceberg_tables_collection.create
        create_fn(iceberg_table=iceberg_table, mode=mode, copy_grants=copy_grants)

        _mock_api.create_snowflake_managed_iceberg_table.assert_called_once_with(
            _mock_iceberg_tables_collection.schema.database.name,
            _mock_iceberg_tables_collection.schema.name,
            iceberg_table,
            create_mode=mode,
            async_req=is_async,
            copy_grants=copy_grants,
        )



import pytest

from snowflake.core.database_role import ContainingScope, Securable
from snowflake.core.exceptions import NotFoundError


pytestmark = pytest.mark.min_sf_ver("8.39.0")


@pytest.mark.use_accountadmin
def test_revoke_role(
    database_roles_in_test_database, test_database_role_name, test_database_role_name_2, test_database_for_grant_name
):
    database_roles_in_test_database[test_database_role_name].grant_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name_2, database=test_database_for_grant_name)
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 2

    database_roles_in_test_database[test_database_role_name].revoke_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name_2, database=test_database_for_grant_name)
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 1


@pytest.mark.use_accountadmin
def test_revoke_privileges(
    database_roles_in_test_database,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
):
    # grant on a table
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 2

    database_roles_in_test_database[test_database_role_name].revoke_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 1

    # check we can revoke again

    database_roles_in_test_database[test_database_role_name].revoke_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    # revoke on a random table
    with pytest.raises(NotFoundError):
        database_roles_in_test_database[test_database_role_name].revoke_privileges(
            privileges=["SELECT"],
            securable_type="TABLE",
            securable=Securable(
                database=test_database_for_grant_name, schema=test_schema_for_grant_name, name="RANDOM_TABLE"
            ),
        )


@pytest.mark.use_accountadmin
def test_revoke_privileges_on_all(
    database_roles_in_test_database, test_database_role_name, test_database_for_grant_name
):
    # all schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_privileges_on_all(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    # there would be more than one schema in the database
    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) >= 2

    database_roles_in_test_database[test_database_role_name].revoke_privileges_on_all(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 1


@pytest.mark.use_accountadmin
def test_revoke_future_privileges(
    database_roles_in_test_database, test_database_role_name, test_database_for_grant_name
):
    # all future schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_future_privileges(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_future_grants_to())) == 1

    database_roles_in_test_database[test_database_role_name].revoke_future_privileges(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_future_grants_to())) == 0


@pytest.mark.use_accountadmin
def test_revoke_grant_option_for_privileges(
    database_roles_in_test_database,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
):
    # grant on a table
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
        grant_option=True,
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grants_list) == 2
    for grant in grants_list:
        if not (grant.securable_type == "DATABASE" and grant.privileges == ["USAGE"]):
            assert grant.grant_option

    database_roles_in_test_database[test_database_role_name].revoke_grant_option_for_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grants_list) == 2
    for grant in grants_list:
        assert not grant.grant_option

    # revoke grant option on a grant which does not have grant option
    database_roles_in_test_database[test_database_role_name].revoke_grant_option_for_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grants_list) == 2
    for grant in grants_list:
        assert not grant.grant_option


@pytest.mark.use_accountadmin
def test_revoke_grant_option_for_privileges_on_all(
    database_roles_in_test_database, test_database_role_name, test_database_for_grant_name
):
    # all schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_privileges_on_all(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
        grant_option=True,
    )

    # there would be more than one schema in the database
    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    number_of_previleges = len(grants_list)
    assert number_of_previleges >= 2
    for grant in grants_list:
        if not (grant.securable_type == "DATABASE" and grant.privileges == ["USAGE"]):
            assert grant.grant_option

    database_roles_in_test_database[test_database_role_name].revoke_grant_option_for_privileges_on_all(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grants_list) == number_of_previleges
    for grant in grants_list:
        assert not grant.grant_option


@pytest.mark.use_accountadmin
def test_revoke_grant_option_for_future_privileges(
    database_roles_in_test_database, test_database_role_name, test_database_for_grant_name
):
    # all future schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_future_privileges(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
        grant_option=True,
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_future_grants_to())
    assert len(grants_list) == 1
    for grant in grants_list:
        assert grant.grant_option

    database_roles_in_test_database[test_database_role_name].revoke_grant_option_for_future_privileges(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    grants_list = list(database_roles_in_test_database[test_database_role_name].iter_future_grants_to())
    assert len(grants_list) == 1
    for grant in grants_list:
        assert not grant.grant_option



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.schema import Schema
from snowflake.core.user_defined_function import PythonFunction, ReturnDataType, UserDefinedFunction
from tests.utils import random_string


pytestmark = [pytest.mark.min_sf_ver("8.38.0"), pytest.mark.usefixtures("anaconda_package_available")]


def test_rename(user_defined_functions):
    user_defined_function_name = random_string(10, "test_original_user_defined_function_")
    user_defined_function_other_name = random_string(10, "test_other_user_defined_function_")

    user_defined_function_handle = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnDataType(datatype="VARIANT"),
            language_config=PythonFunction(runtime_version="3.13", packages=[], handler="udf"),
            body="""
def udf():
    return {"key": "value"}
            """,
        )
    )

    try:
        user_defined_function_handle.fetch()

        user_defined_function_handle.rename(user_defined_function_other_name)

        assert user_defined_function_handle.fetch().name.upper() == user_defined_function_other_name.upper()

        with pytest.raises(NotFoundError):
            user_defined_functions[f"{user_defined_function_name}()"].fetch()

        with pytest.raises(NotFoundError):
            user_defined_functions[f"{user_defined_function_name}()"].rename(user_defined_function_other_name)

        user_defined_functions[f"{user_defined_function_name}()"].rename(
            user_defined_function_other_name, if_exists=True
        )
        user_defined_functions[f"{user_defined_function_other_name}()"].fetch()
    finally:
        # This UDF has to be existed
        user_defined_function_handle.drop()


def test_relocate_schema(user_defined_functions, temp_schema):
    user_defined_function_name = random_string(10, "test_original_user_defined_function_")
    user_defined_function_other_name = random_string(10, "test_other_user_defined_function_")

    user_defined_function_handle = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnDataType(datatype="VARIANT"),
            language_config=PythonFunction(runtime_version="3.13", packages=[], handler="udf"),
            body="""
def udf():
    return {"key": "value"}
            """,
        )
    )

    try:
        user_defined_function_handle.fetch()

        user_defined_function_handle.rename(user_defined_function_other_name, target_schema=temp_schema.name)

        assert user_defined_function_handle.fetch().name.upper() == user_defined_function_other_name.upper()
        assert user_defined_function_handle.fetch().schema_name.upper() == temp_schema.name.upper()
        assert user_defined_function_handle.collection.schema.name.upper() == temp_schema.name.upper()

        with pytest.raises(NotFoundError):
            user_defined_functions[f"{user_defined_function_name}()"].fetch()

        with pytest.raises(NotFoundError):
            user_defined_functions[f"{user_defined_function_name}()"].rename(user_defined_function_other_name)

        user_defined_functions[f"{user_defined_function_name}()"].rename(
            user_defined_function_other_name, if_exists=True
        )
        temp_schema.user_defined_functions[f"{user_defined_function_other_name}()"].fetch()
    finally:
        # This UDF has to be existed
        user_defined_function_handle.drop()


def test_relocate_database(user_defined_functions, temp_db):
    temp_schema_name = random_string(10, "test_temp_schema_")

    created_schema = temp_db.schemas.create(Schema(name=temp_schema_name))

    try:
        user_defined_function_name = random_string(10, "test_original_user_defined_function_")
        user_defined_function_other_name = random_string(10, "test_other_user_defined_function_")

        user_defined_function_handle = user_defined_functions.create(
            UserDefinedFunction(
                name=user_defined_function_name,
                arguments=[],
                return_type=ReturnDataType(datatype="VARIANT"),
                language_config=PythonFunction(runtime_version="3.13", packages=[], handler="udf"),
                body="""
def udf():
    return {"key": "value"}
                """,
            )
        )

        try:
            user_defined_function_handle.fetch()

            user_defined_function_handle.rename(
                user_defined_function_other_name, target_schema=created_schema.name, target_database=temp_db.name
            )

            assert user_defined_function_handle.fetch().name.upper() == user_defined_function_other_name.upper()
            assert user_defined_function_handle.fetch().schema_name.upper() == created_schema.name.upper()
            assert user_defined_function_handle.fetch().database_name.upper() == temp_db.name.upper()
            assert user_defined_function_handle.collection.database.name.upper() == temp_db.name.upper()
            assert user_defined_function_handle.collection.schema.name.upper() == created_schema.name.upper()

            with pytest.raises(NotFoundError):
                user_defined_functions[f"{user_defined_function_name}()"].fetch()

            with pytest.raises(NotFoundError):
                user_defined_functions[f"{user_defined_function_name}()"].rename(user_defined_function_other_name)

            user_defined_functions[f"{user_defined_function_name}()"].rename(
                user_defined_function_other_name, if_exists=True
            )
            created_schema.user_defined_functions[f"{user_defined_function_other_name}()"].fetch()
        finally:
            # This UDF has to be existed
            user_defined_function_handle.drop()
    finally:
        created_schema.drop()



import pytest

from snowflake.core.streamlit import (
    AddVersionStreamlitRequest,
    AddVersionStreamlitRequestVersion,
    Streamlit,
)
from tests.integ.utils import random_string


@pytest.mark.min_sf_ver("9.38.0")
def test_add_version_streamlit(streamlits, streamlit_stage_with_file, warehouse, streamlit_main_file):
    name = random_string(8, "test_add_version_streamlit_")

    st = Streamlit(
        name=name,
        query_warehouse=warehouse.name,
        source_location=f"@{streamlit_stage_with_file.name}",
        main_file=streamlit_main_file,
    )

    ref = streamlits.create(st)
    try:
        version_name = random_string(8, "version_")
        version = AddVersionStreamlitRequestVersion(
            name=version_name,
            comment="Test version for integration testing",
            if_not_exists=True,
        )

        add_version_request = AddVersionStreamlitRequest(
            source_location=f"@{streamlit_stage_with_file.name}",
            version=version,
        )

        ref.add_version(add_version_request)

        fetched = ref.fetch()
        assert fetched.name.upper() == name.upper()
        assert fetched.last_version_details is not None
        assert fetched.last_version_details.name is not None
        assert fetched.last_version_details.alias.upper() == version_name.upper()
        assert fetched.last_version_details.source_location_uri.upper() == f"@{streamlit_stage_with_file.name}".upper()

    finally:
        ref.drop()



import copy

import pytest

from snowflake.core.exceptions import NotFoundError
from tests.integ.utils import random_string

from .conftest import test_network_rule_template


def test_drop(network_rules):
    prefix = "test_network_rule_drop_"
    pre_create_count = len(list(network_rules.iter(like=prefix + "%")))

    name = random_string(10, prefix)
    network_rule_handle = network_rules[name]

    network_rule = copy.deepcopy(test_network_rule_template)
    network_rule.name = name
    network_rules.create(network_rule)

    created_count = len(list(network_rules.iter(like=prefix + "%")))
    assert pre_create_count + 1 == created_count

    network_rule_handle.drop()

    after_drop_count = len(list(network_rules.iter(like=prefix + "%")))
    assert pre_create_count == after_drop_count

    with pytest.raises(NotFoundError):
        network_rule_handle.drop()
        network_rule_handle.drop(if_exists=False)

    network_rule_handle.drop(if_exists=True)



import enum
import functools
import os
import warnings

from re import compile
from typing import TYPE_CHECKING, Callable, Optional, TypeVar, Union

from typing_extensions import ParamSpec

from snowflake.connector import ProgrammingError
from snowflake.connector.description import OPERATING_SYSTEM, PLATFORM
from snowflake.core.exceptions import FileGetError, FilePutError


if TYPE_CHECKING:
    from snowflake.core._root import Root


# The following code is copied from snowpark's code /snowflake/snowpark/_internal/utils.py to avoid being broken
# when snowpark changes the code.
# We'll need to move the code to a common place.
# Another solution is to move snowpark to the mono repo so the merge gate will find the breaking changes.
# To address later.

EMPTY_STRING = ""
DOUBLE_QUOTE = '"'
ALREADY_QUOTED = compile('^(".+")$')
UNQUOTED_CASE_INSENSITIVE = compile("^([_A-Za-z]+[_A-Za-z0-9$]*)$")
# https://docs.snowflake.com/en/sql-reference/identifiers-syntax.html
SNOWFLAKE_UNQUOTED_ID_PATTERN = r"([a-zA-Z_][\w\$]{0,255})"
SNOWFLAKE_QUOTED_ID_PATTERN = '("([^"]|""){1,255}")'
SNOWFLAKE_ID_PATTERN = f"({SNOWFLAKE_UNQUOTED_ID_PATTERN}|{SNOWFLAKE_QUOTED_ID_PATTERN})"
SNOWFLAKE_OBJECT_RE_PATTERN = compile(
    f"^(({SNOWFLAKE_ID_PATTERN}\\.){{0,2}}|({SNOWFLAKE_ID_PATTERN}\\.\\.)){SNOWFLAKE_ID_PATTERN}$"
)


def normalize_datatype(datatype: str) -> str:
    """Convert equivalent datatypes.

    These are according to https://docs.snowflake.com/en/sql-reference/intro-summary-data-types
    All string types to varchar, and all numeric types to number and float.
    """
    datatype = datatype.upper()
    datatype = datatype.replace(" ", "")
    if datatype in ("INT", "INTEGER", "BIGINT", "SMALLINT", "TINYINT", "BYTEINT", "NUMBER"):
        return "NUMBER(38,0)"
    if datatype in ("DOUBLE", "DOUBLEPRECISION", "REAL"):
        return "FLOAT"
    if datatype in ("STRING", "TEXT", "VARCHAR"):
        return "VARCHAR(16777216)"
    if datatype in ("CHAR", "CHARACTER"):
        return "VARCHAR(1)"
    if datatype in ("VARBINARY"):
        return "BINARY"
    datatype = (
        datatype.replace("DECIMAL", "NUMBER")
        .replace("NUMERIC", "NUMBER")
        .replace("STRING", "VARCHAR")
        .replace("TEXT", "VARCHAR")
    )
    return datatype


def validate_object_name(name: str) -> None:
    if not SNOWFLAKE_OBJECT_RE_PATTERN.match(name):
        raise ValueError(f"The object name '{name}' is invalid.")


def is_running_inside_stored_procedure() -> bool:
    """
    Check if snowpy is running inside a stored procedure.

    Returns:
        bool: True if snowpy is running inside a stored procedure, False otherwise.
    """
    return PLATFORM == "XP"


def validate_quoted_name(name: str) -> str:
    if DOUBLE_QUOTE in name[1:-1].replace(DOUBLE_QUOTE + DOUBLE_QUOTE, EMPTY_STRING):
        raise ValueError(
            f"Invalid Identifier {name}. "
            f"The inside double quotes need to be escaped when the name itself is double quoted."
        )
    else:
        return name


def escape_quotes(unescaped: str) -> str:
    return unescaped.replace(DOUBLE_QUOTE, DOUBLE_QUOTE + DOUBLE_QUOTE)


def normalize_name(name: str) -> str:
    if ALREADY_QUOTED.match(name):
        return validate_quoted_name(name)
    elif UNQUOTED_CASE_INSENSITIVE.match(name):
        return escape_quotes(name.upper())
    else:
        return DOUBLE_QUOTE + escape_quotes(name) + DOUBLE_QUOTE


def normalize_and_unquote_name(name: str) -> str:
    return unquote_name(normalize_name(name))


def unquote_name(name: str) -> str:
    if len(name) > 1 and name[0] == name[-1] == '"':
        return name[1:-1]
    return name


P = ParamSpec("P")
R = TypeVar("R")


def deprecated(alternative: Optional[str] = None) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        """Mark methods as deprecated with a warning message."""
        deprecation_message = f"The :func:`{func.__name__}` method is deprecated;"
        if alternative:
            deprecation_message += f" use :func:`{alternative}` instead."

        @functools.wraps(func)
        def deprecate_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            warnings.warn(deprecation_message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        if deprecate_wrapper.__doc__ is None:
            deprecate_wrapper.__doc__ = deprecation_message
        else:
            deprecate_wrapper.__doc__ = deprecation_message + "\n\n" + deprecate_wrapper.__doc__

        return deprecate_wrapper

    return decorator


@enum.unique
class TelemetryField(enum.Enum):
    # Copied over from snowflake.snowpark
    # constants
    MESSAGE = "message"
    NAME = "name"
    ERROR_CODE = "error_code"
    STACK_TRACE = "stack_trace"
    # Types of telemetry
    TYPE_PERFORMANCE_DATA = "snowpark_performance_data"
    TYPE_FUNCTION_USAGE = "snowpark_function_usage"
    TYPE_SESSION_CREATED = "snowpark_session_created"
    TYPE_SQL_SIMPLIFIER_ENABLED = "snowpark_sql_simplifier_enabled"
    TYPE_CTE_OPTIMIZATION_ENABLED = "snowpark_cte_optimization_enabled"
    # telemetry for optimization that eliminates the extra cast expression generated for expressions
    TYPE_ELIMINATE_NUMERIC_SQL_VALUE_CAST_ENABLED = "snowpark_eliminate_numeric_sql_value_cast_enabled"
    TYPE_AUTO_CLEAN_UP_TEMP_TABLE_ENABLED = "snowpark_auto_clean_up_temp_table_enabled"
    TYPE_ERROR = "snowpark_error"
    # Message keys for telemetry
    KEY_START_TIME = "start_time"
    KEY_DURATION = "duration"
    KEY_FUNC_NAME = "func_name"
    KEY_MSG = "msg"
    KEY_ERROR_MSG = "error_msg"
    KEY_VERSION = "version"
    KEY_PYTHON_VERSION = "python_version"
    KEY_CLIENT_LANGUAGE = "client_language"
    KEY_OS = "operating_system"
    KEY_DATA = "data"
    KEY_CATEGORY = "category"
    KEY_CREATED_BY_SNOWPARK = "created_by_snowpark"
    KEY_API_CALLS = "api_calls"
    KEY_SFQIDS = "sfqids"
    KEY_SUBCALLS = "subcalls"
    KEY_CI_ENVIRONMENT_TYPE = "ci_environment_type"
    # function categories
    FUNC_CAT_ACTION = "action"
    FUNC_CAT_USAGE = "usage"
    FUNC_CAT_JOIN = "join"
    FUNC_CAT_COPY = "copy"
    FUNC_CAT_CREATE = "create"
    # performance categories
    PERF_CAT_UPLOAD_FILE = "upload_file"
    # sql simplifier
    SESSION_ID = "session_id"
    SQL_SIMPLIFIER_ENABLED = "sql_simplifier_enabled"
    CTE_OPTIMIZATION_ENABLED = "cte_optimization_enabled"
    # dataframe query stats
    QUERY_PLAN_HEIGHT = "query_plan_height"
    QUERY_PLAN_NUM_DUPLICATE_NODES = "query_plan_num_duplicate_nodes"
    QUERY_PLAN_COMPLEXITY = "query_plan_complexity"


STAGE_PREFIX = "@"
SNOWURL_PREFIX = "snow://"
SNOWFLAKE_PATH_PREFIXES = [STAGE_PREFIX, SNOWURL_PREFIX]


def is_single_quoted(name: str) -> bool:
    return name.startswith("'") and name.endswith("'") and len(name) >= 2


def normalize_path(path: str, is_local: bool) -> str:
    """Get a normalized path of a local file or remote stage location for PUT/GET commands.

    If there are any special characters including spaces in the path, it needs to be
    quoted with single quote. For example, 'file:///tmp/load data' for a path containing
    a directory named "load data". Therefore, if `path` is already wrapped by single quotes,
    we do nothing.
    """
    prefixes = ["file://"] if is_local else SNOWFLAKE_PATH_PREFIXES
    if is_single_quoted(path):
        return path
    if is_local and OPERATING_SYSTEM == "Windows":
        path = path.replace("\\", "/")
    path = path.strip().replace("'", "\\'")
    if not any(path.startswith(prefix) for prefix in prefixes):
        path = f"{prefixes[0]}{path}"
    return f"'{path}'"


def get_local_file_path(file: str) -> str:
    trim_file = file.strip()
    if is_single_quoted(trim_file):
        trim_file = trim_file[1:-1]  # remove the pair of single quotes
    if trim_file.startswith("file://"):
        return trim_file[7:]  # remove "file://"
    return trim_file


def get_file(
    root: "Root", stage_location: str, target_directory: str, *, parallel: int = 10, pattern: Optional[str] = None
) -> None:
    options: dict[str, Union[int, str]] = {"parallel": parallel}
    if pattern is not None:
        if not is_single_quoted(pattern):
            pattern_escape_single_quote = pattern.replace("'", "\\'")
            pattern = f"'{pattern_escape_single_quote}'"  # snowflake pattern is a string with single quote
        options["pattern"] = pattern

    if is_running_inside_stored_procedure():  # pragma: no cover
        try:
            cur = root.connection.cursor()
            cur._download(stage_location, target_directory, options)
            cur.fetchall()  # Make sure no errors were raised
        except ProgrammingError as pe:
            raise FileGetError(pe.msg) from None
    else:
        os.makedirs(get_local_file_path(target_directory), exist_ok=True)
        with root.connection.cursor() as cur:
            cur.execute(
                "GET /* snowflake.core.stage.StageResource.get */ "
                + normalize_path(stage_location, is_local=False)
                + " "
                + normalize_path(target_directory, is_local=True)
                + " "
                + f"PARALLEL = {parallel}"
                + (f" PATTERN = {pattern};" if pattern is not None else ";")
            )
            cur.fetchall()  # Make sure no errors were raised


def put_file(
    root: "Root",
    local_file_name: str,
    stage_location: str,
    *,
    parallel: int = 4,
    auto_compress: bool = True,
    source_compression: str = "AUTO_DETECT",
    overwrite: bool = False,
) -> None:
    if is_running_inside_stored_procedure():  # pragma: no cover
        try:
            cur = root.connection.cursor()
            cur._upload(
                local_file_name,
                stage_location,
                {
                    "parallel": parallel,
                    "source_compression": source_compression,
                    "auto_compress": auto_compress,
                    "overwrite": overwrite,
                },
            )
            cur.fetchall()  # Make sure no errors were raised
        except ProgrammingError as pe:
            raise FilePutError(pe.msg) from None
    else:
        with root.connection.cursor() as cur:
            cur.execute(
                "PUT /* snowflake.core.stage.StageResource.put */ "
                + normalize_path(local_file_name, is_local=True)
                + " "
                + normalize_path(stage_location, is_local=False)
                + " "
                + f"PARALLEL = {parallel} "
                + f"AUTO_COMPRESS = {auto_compress} "
                + f"SOURCE_COMPRESSION = {source_compression} "
                + f"OVERWRITE = {overwrite};"
            )
            cur.fetchall()  # Make sure no errors were raised



"""Manages Snowflake tag."""

from ._generated.models import (
    Tag,
)
from ._tag import TagCollection, TagResource, TagValue


__all__ = [
    "TagResource",
    "TagCollection",
    "TagValue",
    "Tag",
]



"""High level, client-side representation of task graphs.

This set of higher-level classes provides a more convenient way to create, deploy, and manage task graphs
than the lower-level Task APIs in `snowflake.core.task`. Task graphs are directed acyclic graphs (DAG) of tasks.

Example 1: Create a task graph that has two Tasks.
    >>> from snowflake.snowpark.functions import sum as sum_
    >>> from snowflake.core.task import StoredProcedureCall
    >>> from snowflake.core.task.dagv1 import DAG, DAGTask, DAGOperation
    >>> def dosomething(session: Session) -> None:
    ...     df = session.table("target")
    ...     df.group_by("a").agg(sum_("b")).save_as_table("agg_table")
    >>> with DAG("my_dag", schedule=timedelta(days=1)) as dag:
    ...     # Create a task that runs some SQL.
    ...     dag_task1 = DAGTask(
    ...         "dagtask1",
    ...         "MERGE INTO target USING source_stream WHEN MATCHED THEN UPDATE SET target.v = source_stream.v")
    ...     # Create a task that runs a Python function.
    ...     dag_task2 = DAGTask(
    ...         StoredProcedureCall(
    ...             dosomething, stage_location="@mystage",
    ...             packages=["snowflake-snowpark-python"]
    ...         ),
    ...         warehouse="test_warehouse")
    ...     )
    >>> # Shift right and left operators can specify task relationships.
    >>> dag_task1 >> dag_task2
    >>> schema = root.databases["MYDB"].schemas["MYSCHEMA"]
    >>> dag_op = DAGOperation(schema)
    >>> dag_op.deploy(dag)

Example 2: Create a task graph that uses Cron, Branch, and function return value as Task return value
    >>> from snowflake.snowpark import Session
    >>> from snowflake.core import Root
    >>> from snowflake.core._common import CreateMode
    >>> from snowflake.core.task import Cron
    >>> from snowflake.core.task.dagv1 import DAG, DAGTask, DAGOperation, DAGTaskBranch
    >>> session = Session.builder.create()
    >>> test_stage = "mystage"
    >>> test_dag = "mydag"
    >>> test_db = "mydb"
    >>> test_schema = "public"
    >>> test_warehouse = "testwh_python"
    >>> root = Root(session)
    >>> schema = root.databases[test_db].schemas[test_schema]
    >>> def task_handler1(session: Session) -> None:
    ...     pass  # do something
    >>> def task_handler2(session: Session) -> None:
    ...     pass  # do something
    >>> def task_handler3(session: Session) -> None:
    ...     pass  # do something
    >>> def task_branch_handler(session: Session) -> str:
    ...     # do something
    ...     return "task3"
    >>> try:
    ...     with DAG(
    ...         test_dag,
    ...         schedule=Cron("10 * * * *", "America/Los_Angeles"),
    ...         stage_location=test_stage,
    ...         packages=["snowflake-snowpark-python"],
    ...         warehouse=test_warehouse,
    ...         use_func_return_value=True,
    ...     ) as dag:
    ...         task1 = DAGTask(
    ...             "task1",
    ...             task_handler1,
    ...         )
    ...         task1_branch = DAGTaskBranch("task1_branch", task_branch_handler, warehouse=test_warehouse)
    ...         task2 = DAGTask("task2", task_handler2)
    ...         task1 >> task1_branch
    ...         task1_branch >> [task2, task_handler3]  # after >> you can use a DAGTask or a function.
    ...     op = DAGOperation(schema)
    ...     op.deploy(dag, mode=CreateMode.or_replace)
    >>> finally:
    ...     session.close()
"""

import json
import logging
import warnings

from collections import deque
from collections.abc import Iterable, Iterator
from contextlib import suppress
from datetime import datetime, timedelta
from functools import wraps
from types import ModuleType, TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Protocol, Union, runtime_checkable

from typing_extensions import Self

from snowflake.core import FQN
from snowflake.core._common import CreateMode
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._internal.utils import deprecated, normalize_name
from snowflake.core.exceptions import InvalidOperationError, NotFoundError
from snowflake.core.schema import SchemaResource
from snowflake.core.task import Cron, StoredProcedureCall, Task
from snowflake.core.task._generated import TaskRun


if TYPE_CHECKING:
    from snowflake.snowpark import Session


@runtime_checkable
class SQLConvertible(Protocol):
    """Protocol for definitions that can be converted to SQL strings."""

    def to_sql(self) -> str:
        """Convert the definition to a SQL string."""
        ...


logger = logging.getLogger(__name__)
_dag_context_stack: deque["DAG"] = deque()


def _get_current_dag() -> Optional["DAG"]:
    if _dag_context_stack:
        return _dag_context_stack[-1]
    return None


def _use_func_return_value(f: Callable[["Session"], Optional[str]]) -> Callable[["Session"], Optional[str]]:
    @wraps(f)
    def wrapper(session: "Session") -> Optional[str]:
        f_return_value = f(session)
        if f_return_value is not None:
            try:
                session.call("system$set_return_value", str(f_return_value))
            except Exception as sse:
                if "Function SYSTEM$SET_RETURN_VALUE must be called from within a task" not in str(sse):
                    raise
        return f_return_value

    return wrapper


def _convert_func_to_task(
    other: Union[
        "DAGTask",
        Iterable[Union["DAGTask", Callable[["Session"], Optional[str]]]],
        Callable[["Session"], Optional[str]],
    ],
) -> Union["DAGTask", Iterable["DAGTask"]]:
    tasks: Union[DAGTask, Iterable[DAGTask]]
    if callable(other):
        tasks = DAGTask(other.__name__, other)
    elif isinstance(other, DAGTask):
        tasks = other
    elif isinstance(other, Iterable):
        tasks = [DAGTask(t.__name__, t) if callable(t) else t for t in other]
    else:
        raise TypeError(f"Expect a DAGTAsk, a function, or a list of them. {type(other)} is provided.")
    return tasks


def _add_condition(full_name: str, successor: "DAGTask") -> None:
    validate_branch = f"SYSTEM$GET_PREDECESSOR_RETURN_VALUE('{normalize_name(full_name)}') = '{successor.name}'"
    if successor.condition:
        successor.condition = f"{validate_branch} and {successor.condition}"
    else:
        successor.condition = validate_branch


class DAG:
    """A graph of tasks composed of a single root task and additional tasks, organized by their dependencies.

    Snowflake doesn't have a first-class task graph entity, so this is a client-side object representation which
    manages Task relationship. A root :class:`Task` and its successors logically form a task graph or DAG
    (Directed Acyclic Graph).  Refer to
    `Task graphs <https://docs.snowflake.com/en/user-guide/tasks-intro#task-graphs>`_.

    When a task graph is deployed, all child tasks are created in Snowflake. A dummy Task is created as the root.
    A task's predecessor is the dummy task if it's added to the task graph with no other predecessors.

    Example:
        >>> dag = DAG(
        ...     "TEST_DAG",
        ...     schedule=timedelta(minutes=10),
        ...     use_func_return_value=True,
        ...     warehouse="TESTWH_DAG",
        ...     packages=["snowflake-snowpark-python"],
        ...     stage_location="@TESTDB_DAG.TESTSCHEMA_DAG.TEST_STAGE_DAG",
        ... )
        >>> def task1(session: Session) -> None:
        ...     session.sql("select 'task1'").collect()
        >>> def task2(session: Session) -> None:
        ...     session.sql("select 'task2'").collect()
        >>> def cond(session: Session) -> str:
        ...     return "TASK1"
        >>> with dag:
        ...     task1 = DAGTask("TASK1", definition=task1, warehouse="TESTWH_DAG")
        ...     task2 = DAGTask("TASK2", definition=task2, warehouse="TESTWH_DAG")
        ...     condition = DAGTaskBranch("COND", definition=cond, warehouse="TESTWH_DAG")
        ...     condition >> [task1, task2]
        >>> dag_op = DAGOperation(schema)
        >>> dag_op.deploy(dag, mode="orReplace")
        >>> dag_op.run(dag)
        Note:
            When defining a task branch handler, simply return the task name you want to jump to. The task name is
            case-sensitive, and it has to match the name property in DAGTask. For exmaple, in above sample code, return
            'TASK1' instead of 'TEST_DAG$TASK1', 'task1' or 'Task1' will not be considered as a exact match.

    Refer to :class:`snowflake.core.task.Task` for the details of each property.
    """

    def __init__(
        self,
        name: str,
        *,
        schedule: Optional[Union[Cron, timedelta]] = None,
        warehouse: Optional[str] = None,
        user_task_managed_initial_warehouse_size: Optional[str] = None,
        error_integration: Optional[str] = None,
        comment: Optional[str] = None,
        task_auto_retry_attempts: Optional[int] = None,
        allow_overlapping_execution: Optional[bool] = None,
        user_task_timeout_ms: Optional[int] = None,
        suspend_task_after_num_failures: Optional[int] = None,
        config: Optional[dict[str, Any]] = None,
        session_parameters: Optional[dict[str, Any]] = None,
        stage_location: Optional[str] = None,
        imports: Optional[list[Union[str, tuple[str, str]]]] = None,
        packages: Optional[list[Union[str, ModuleType]]] = None,
        use_func_return_value: bool = False,
    ) -> None:
        #: Name of the task graph and the dummy root task.
        self.name = name
        #: Refer to :attr:`snowflake.core.task.Task.warehouse`.
        self.warehouse = warehouse
        #: comment of the task graph.
        self.comment = comment
        #: Schedule of the task graph. Refer to :attr:`snowflake.core.task.Task.schedule`.
        self.schedule = schedule
        #: Refer to :attr:`snowflake.core.task.Task.error_integration`.
        self.error_integration = error_integration
        #: Refer to :attr:`snowflake.core.task.Task.allow_overlapping_execution`.
        self.allow_overlapping_execution = allow_overlapping_execution
        #: Refer to :attr:`snowflake.core.task.Task.user_task_timeout_ms`.
        self.user_task_timeout_ms = user_task_timeout_ms
        #: Refer to :attr:`snowflake.core.task.Task.suspend_task_after_num_failures`.
        self.suspend_task_after_num_failures = suspend_task_after_num_failures
        #: Refer to :attr:`snowflake.core.task.Task.config`.
        self.config = config
        #: Refer to :attr:`snowflake.core.task.Task.session_parameters`.
        self.session_parameters = session_parameters
        self._tasks: dict[str, DAGTask] = dict()
        self._task_list: Optional[list[DAGTask]] = None  # as a cache for property tasks

        #: The default stage location where this task graph's tasks code will be stored
        #: if creating the tasks from Python functions.
        self.stage_location = stage_location
        #: The default imports for all tasks of this task graph if creating the tasks from Python functions.
        self.imports = imports
        #: The default packages for the tasks of this task graph if creating the tasks from Python functions.
        self.packages = packages

        #: Use the Python function's return value as Task return value if ``use_func_return_value`` is True.
        #: Default False.
        self.use_func_return_value = use_func_return_value

        # name of the finalizer task, this can be none because it is not necessary to have a finalizer task
        # remember that there can only be one finalizer task
        self._finalizer_task: Optional[str] = None

        #: Refer to :attr:`snowflake.core.task.Task.task_auto_retry_attempts`.
        self.task_auto_retry_attempts = task_auto_retry_attempts

        if user_task_managed_initial_warehouse_size is not None:
            warnings.warn(
                "Providing user_task_managed_initial_warehouse_size for a dummy root Task is deprecated and "
                "will be removed in the next major version. You can specify user_task_managed_initial_warehouse_size "
                "for the child Tasks instead.",
                category=DeprecationWarning,
                stacklevel=1,
            )

    def _to_low_level_task(self) -> Task:
        return Task(
            name=f"{self.name}",
            definition="select 'dag dummy root'",
            schedule=self.schedule,
            warehouse=self.warehouse,
            error_integration=self.error_integration,
            comment=self.comment,
            task_auto_retry_attempts=self.task_auto_retry_attempts,
            allow_overlapping_execution=self.allow_overlapping_execution,
            user_task_timeout_ms=self.user_task_timeout_ms,
            suspend_task_after_num_failures=self.suspend_task_after_num_failures,
            session_parameters=self.session_parameters,
            config=self.config,
        )

    def add_task(self, task: "DAGTask") -> None:
        """Add a child task to this task graph.

        Parameters
        __________
        task: DAGTask
            The child task to be added to this task graph.

        Examples
        ________
        Add a task to previously created DAG

        >>> child_task = DagTask("child_task", "select 'child_task'", warehouse="test_warehouse")
        >>> dag.add_task(child_task)
        )
        """
        if task.dag is not self:
            raise ValueError("Cannot move a task to a different task graph")
        self._tasks[task.name] = task
        self._task_list = None

    def get_task(self, task_name: str) -> Optional["DAGTask"]:
        """Get a child task from this task graph based on task name.

        Parameters
        __________
        task_name: str
            The name of the task to be retrieved from this task graph.

        Examples
        ________
        Get a task from previously created DAG

        >>> task = dag.get_task("child_task")
        """
        return self._tasks.get(task_name)

    def get_finalizer_task(self) -> Optional["DAGTask"]:
        """Get the finalizer task for the dag.

        Examples
        ________
        Get the finalizer task from previously created DAG:

        >>> finalizer_task = dag.get_finalizer_task()
        """
        if self._finalizer_task is not None:
            return self._tasks.get(str(self._finalizer_task))
        return None

    @property
    def tasks(self) -> list["DAGTask"]:
        """Returns a list of tasks this task graph has."""
        if self._task_list is None:
            self._task_list = list(self._tasks.values())
        return self._task_list

    def __repr__(self) -> str:
        return f"<DAG: {self.name}>"

    def __contains__(self, task: Union["DAGTask", str]) -> bool:
        task_name = task.name if isinstance(task, DAGTask) else task
        return task_name in self._tasks

    def __enter__(self) -> Self:
        _dag_context_stack.append(self)
        return self

    def __exit__(
        self, exc_type: Optional[type[BaseException]], exc: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> None:
        _dag_context_stack.pop()


class DAGTask:
    """Represents a child Task of a task graph.

    A child task has a subset of properties of :class:`snowflake.core.task.Task`. For instance, it doesn't
    have ``schedule`` because it's triggered after the predecessors are finished.

    Parameters
    __________
    is_finalizer: bool, optional
        Whether this child task is a finalizer task for the DAG.
    is_serverless: bool
        Whether the child task is serverless. If set, no warehouse can be specified for the task. If not set
        and no warehouse is specified for the child Task, then the warehouse setting is inherited from the root Task.
        This means that if both the child Task and its DAG do not have a warehouse specified then the child Task will
        be created as serverless even if is_serverless is set to False. Default is False.
    dag: DAG, optional
        The DAG this child Task belong to. The value is optional if the task is created within a DAG context,
        otherwise it must be provided.

    Refer to :class:`snowflake.core.task.Task` for the details of the remaining properties.
    """

    def __init__(
        self,
        name: str,
        definition: Union[str, Callable[..., Any], StoredProcedureCall, SQLConvertible],
        *,
        condition: Optional[str] = None,
        warehouse: Optional[str] = None,
        session_parameters: Optional[dict[str, Any]] = None,
        user_task_managed_initial_warehouse_size: Optional[str] = None,
        target_completion_interval: Optional[timedelta] = None,
        serverless_task_min_statement_size: Optional[str] = None,
        serverless_task_max_statement_size: Optional[str] = None,
        user_task_timeout_ms: Optional[int] = None,
        error_integration: Optional[str] = None,
        comment: Optional[str] = None,
        is_finalizer: Optional[bool] = None,
        is_serverless: bool = False,
        dag: Optional[DAG] = None,
    ) -> None:
        self.name = name  #: Name of the Task.
        self.definition = definition  #: refer to :attr:`snowflake.core.task.Task.definition`.
        self.condition = condition  #: refer to :attr:`snowflake.core.task.Task.condition`.
        self.session_parameters = session_parameters  #: refer to :attr:`snowflake.core.task.Task.session_parameter`.
        self.comment = comment  #: comment for the task.
        # refer to :attr:`snowflake.core.task.Task.user_task_managed_initial_warehouse_size`.
        self.user_task_managed_initial_warehouse_size = user_task_managed_initial_warehouse_size
        #: refer to :attr:`snowflake.core.task.Task.target_completion_interval`.
        self.target_completion_interval = target_completion_interval
        #: refer to :attr:`snowflake.core.task.Task.serverless_task_min_statement_size`.
        self.serverless_task_min_statement_size = serverless_task_min_statement_size
        #: refer to :attr:`snowflake.core.task.Task.serverless_task_max_statement_size`.
        self.serverless_task_max_statement_size = serverless_task_max_statement_size
        self.user_task_timeout_ms = user_task_timeout_ms  #: refer to :attr:`snowflake.core.task.user_task_timeout_ms`
        dag = dag or _get_current_dag()
        if dag is None:
            raise ValueError("Parameter 'dag' must be set when creating a DAGTask outside of a DAG context.")
        if is_serverless and warehouse is not None:
            raise ValueError("Warehouse was provided for a serverless task.")
        self.warehouse = dag.warehouse if (warehouse is None and not is_serverless) else warehouse  # : refer to
        # :attr:`snowflake.core.task.Task.warehouse`.

        if error_integration is not None:
            warnings.warn(
                "error_integration cannot be specified for a child Task. This argument will be removed "
                "in the next major version. Provide the error_integration for the DAG instead.",
                category=DeprecationWarning,
                stacklevel=1,
            )

        self._dag = dag
        # This is for creating a finalizer task for the dag
        self._is_finalizer = is_finalizer
        if is_finalizer:
            if self._dag._finalizer_task is None:
                self._dag._finalizer_task = self.name
            else:
                raise InvalidOperationError(
                    f"Finalizer task for this Dag {self.name} already exists named: {self._dag._finalizer_task}"
                )
        self._dag.add_task(self)
        self._predecessors: set[DAGTask] = set()

    @property
    def full_name(self) -> str:
        """The full name of the child task."""
        return f"{self.dag.name}${self.name}"

    @property
    def dag(self) -> DAG:
        """Return the :class:`DAG` of this task."""
        return self._dag

    def add_predecessors(
        self,
        other: Union[
            "DAGTask",
            Iterable[Union["DAGTask", Callable[["Session"], Optional[str]]]],
            Callable[["Session"], Optional[str]],
        ],
    ) -> None:
        """Add a task or a task list to the direct predecessors of the current task.

        Parameters
        __________
        other: Union[DAGTask, Iterable[DAGTask], Callable[["Session"], Optional[str]]]
            The task or task list to be added as the direct predecessors of the current task.

        Examples
        ________
        Add a task to the predecessors of the current task:

        >>> task1 = DAGTask("task1", "select 'task1'")
        >>> task2 = DAGTask("task2", "select 'task2'")
        >>> task1.add_predecessors(task2)
        """
        if self._is_finalizer:
            raise InvalidOperationError(f"Finalizer task {self.name} cannot have any predecessors")
        tasks = _convert_func_to_task(other)
        if isinstance(tasks, DAGTask):
            tasks = [tasks]

        for t in tasks:
            if t.dag is not self.dag:
                raise InvalidOperationError(f"Task {t.name} belongs to a different task graph")
            if t._is_finalizer:
                raise InvalidOperationError(f"Finalizer task {t.name} cannot be predecessor of any task")

        for t in tasks:
            if isinstance(t, DAGTaskBranch):
                _add_condition(t.full_name, self)
        self._predecessors.update(tasks)

    def add_successors(
        self,
        other: Union[
            "DAGTask",
            Iterable[Union["DAGTask", Callable[["Session"], Optional[str]]]],
            Callable[["Session"], Optional[str]],
        ],
    ) -> None:
        """Add a task or a task list to be the direct successor of the current task.

        Parameters
        __________
        other: Union[DAGTask, Iterable[DAGTask], Callable[["Session"], Optional[str]]]
            The task or task list to be added as the direct successors of the current task.

        Examples
        ________
        Add a task to the successors of the current task:

        >>> task1 = DAGTask("task1", "select 'task1'")
        >>> task2 = DAGTask("task2", "select 'task2'")
        >>> task1.add_successors(task2)
        """
        if self._is_finalizer:
            raise InvalidOperationError(f"Finalizer task {self.name} cannot have any successors")
        tasks = _convert_func_to_task(other)
        if isinstance(tasks, DAGTask):
            tasks = [tasks]

        for t in tasks:
            if t.dag is not self.dag:
                raise InvalidOperationError(f"Task {t.name} belongs to a different task graph")
            if t._is_finalizer:
                raise InvalidOperationError(f"Finalizer task {t.name} cannot have any predecessors")

        for t in tasks:
            t.add_predecessors(self)

    @property
    def predecessors(self) -> set["DAGTask"]:
        """Return the predecessors of the Task."""
        return self._predecessors

    def __repr__(self) -> str:
        return f"<DAGTask: {self.name}>"

    def __lshift__(
        self,
        other: Union[
            "DAGTask",
            Iterable[Union["DAGTask", Callable[["Session"], Optional[str]]]],
            Callable[["Session"], Optional[str]],
        ],
    ) -> Union["DAGTask", Iterable["DAGTask"]]:
        """Implement task1 << task2."""
        if self._is_finalizer:
            raise InvalidOperationError(f"Finalizer task {self.name} cannot have any predecessors")
        tasks = _convert_func_to_task(other)
        self.add_predecessors(tasks)
        return tasks

    def __rshift__(
        self,
        other: Union[
            "DAGTask",
            Iterable[Union["DAGTask", Callable[["Session"], Optional[str]]]],
            Callable[["Session"], Optional[str]],
        ],
    ) -> Union["DAGTask", Iterable["DAGTask"]]:
        """Implement task1 >> task2."""
        if self._is_finalizer:
            raise InvalidOperationError(f"Finalizer task {self.name} cannot have any successors")
        tasks = _convert_func_to_task(other)
        self.add_successors(tasks)
        return tasks

    def _to_low_level_task(self) -> Task:
        definition: Union[str, StoredProcedureCall]
        if isinstance(self.definition, SQLConvertible):
            sql = self.definition.to_sql()
            if not isinstance(sql, str):
                raise TypeError(f"to_sql() must return a str, got {type(sql).__name__}")
            definition = sql
        elif isinstance(self.definition, StoredProcedureCall):
            func = self.definition.func
            if isinstance(func, Callable) and self.dag.use_func_return_value:  # type: ignore
                func = _use_func_return_value(func)
            definition = StoredProcedureCall(
                func,
                args=self.definition._args,
                return_type=self.definition._return_type,
                input_types=self.definition._input_types,
                stage_location=self.definition._stage_location or self.dag.stage_location,
                imports=self.definition._imports or self.dag.imports,
                packages=self.definition._packages or self.dag.packages,
            )
        elif isinstance(self.definition, Callable):  # type: ignore
            definition = StoredProcedureCall(
                _use_func_return_value(self.definition) if self.dag.use_func_return_value else self.definition,  # type: ignore
                stage_location=self.dag.stage_location,
                imports=self.dag.imports,
                packages=self.dag.packages,
            )
        else:
            definition = self.definition  # type: ignore

        predecessors = None
        if self.predecessors:
            predecessors = [x.full_name for x in self.predecessors]
        elif not self._is_finalizer:
            predecessors = [self._dag.name]
        return Task(
            name=self.full_name,
            definition=definition,
            condition=self.condition,
            warehouse=self.warehouse,
            user_task_managed_initial_warehouse_size=self.user_task_managed_initial_warehouse_size,
            target_completion_interval=self.target_completion_interval,
            serverless_task_min_statement_size=self.serverless_task_min_statement_size,
            serverless_task_max_statement_size=self.serverless_task_max_statement_size,
            comment=self.comment,
            user_task_timeout_ms=self.user_task_timeout_ms,
            session_parameters=self.session_parameters,
            predecessors=predecessors,
            finalize=self._dag.name if self._is_finalizer else None,
        )


class DAGTaskBranch(DAGTask):
    pass


class DAGRun:
    """Contains the history of a task graph run in Snowflake.

    From https://docs.snowflake.com/en/sql-reference/functions/current_task_graphs#output,
    but tweaked a little bit to fit into task graph context.
    """

    dag_name: str  #: Name of the root task and the task graph.
    database_name: str  #: Name of the database that contains the graph.
    schema_name: str  #: Name of the schema that contains the graph.
    state: str  #: state of the task graph run. One of ["SCHEDULED", "EXECUTING", "SUCCEEDED", "FAILED", "CANCELLED"].
    first_error_task_name: Optional[str]  #: Name of the first task in the graph that returned an error.
    first_error_code: Optional[int]  #: Error code of the error returned by the task named in ``first_error_task_name``.
    first_error_message: Optional[
        str
    ]  #: Error message of the error returned by the task named in ``first_error_task_name``.
    scheduled_time: Optional[datetime]  #: Time when the root task was scheduled to start running.
    query_start_time: Optional[datetime]  #: Time when the query in the root task definition started to run.
    # Time when the root task is next scheduled to start running, assuming the current run of the task graph started
    #  at the scheduled_time completes in time.
    next_scheduled_time: datetime
    run_id: int  #: epoch time in milliseconds of the root task's original scheduled start time.
    graph_version: int  #: Integer identifying the version of the task graph that was run, or is scheduled to be run.

    @classmethod
    def _from_taskrun(cls, taskrun: TaskRun) -> "DAGRun":
        dagrun = cls()
        dagrun.database_name = taskrun.database_name
        dagrun.schema_name = taskrun.schema_name
        dagrun.state = taskrun.state
        dagrun.first_error_task_name = taskrun.first_error_task_name
        dagrun.first_error_code = taskrun.first_error_code
        dagrun.first_error_message = taskrun.first_error_message
        dagrun.scheduled_time = taskrun.scheduled_time
        dagrun.query_start_time = taskrun.query_start_time
        dagrun.next_scheduled_time = taskrun.next_scheduled_time
        dagrun.run_id = taskrun.run_id
        dagrun.graph_version = taskrun.graph_version
        dagrun.dag_name = taskrun.root_task_name
        return dagrun

    def __str__(self) -> str:
        return f"""run_id: {self.run_id}, dag_name: {self.dag_name}, database_name: {self.database_name}, schema_name:
{self.schema_name}, state: {self.state}, first_error_task_name: {self.first_error_task_name},
first_error_code: {self.first_error_code}, first_error_message: {self.first_error_message},
scheduled_time: {self.scheduled_time}, query_start_time: {self.query_start_time},
next_scheduled_time: {self.next_scheduled_time}, graph_version: {self.graph_version}
"""

    def _repr_html_(self) -> str:
        return f"""<table border="1">
<tr><th>Property</th><th>Value</th></tr>
<tr><td>run_id</td><td>{self.run_id}</td></tr>
<tr><td>dag_name</td><td>{self.dag_name}</td></tr>
<tr><td>database_name</td><td>{self.database_name}</td></tr>
<tr><td>schema_name</td><td>{self.schema_name}</td></tr>
<tr><td>state</td><td>{self.state}</td></tr>
<tr><td>first_error_task_name</td><td>{self.first_error_task_name}</td></tr>
<tr><td>first_error_code</td><td>{self.first_error_code}</td></tr>
<tr><td>first_error_message</td><td>{self.first_error_message}</td></tr>
<tr><td>scheduled_time</td><td>{self.scheduled_time}</td></tr>
<tr><td>query_start_time</td><td>{self.query_start_time}</td></tr>
<tr><td>next_scheduled_time</td><td>{self.next_scheduled_time}</td></tr>
<tr><td>graph_version</td><td>{self.graph_version}</td></tr>
</table>
"""


class DAGOperation:
    """APIs to manage task graph child task operations."""

    def __init__(self, schema: SchemaResource) -> None:
        self.schema = schema
        """The schema that the task graph's child tasks will be read from or create into."""

    def iter_dags(self, *, like: str) -> list[str]:
        """Return the task graph names under this schema.

        Parameters
        __________
        like: str
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).

        Examples
        ________
        Get all task graph names under the schema:

        >>> dags = dag_op.iter_dags(like="your-dag-name")
        """
        return [x.name for x in self.schema.tasks.iter(like=like, root_only=True)]

    @api_telemetry
    @deprecated("drop")
    def delete(self, dag: Union[DAG, str]) -> None:
        self.drop(dag=dag)

    @api_telemetry
    def drop(self, dag: Union[DAG, str], *, drop_finalizer: bool = False) -> None:
        """Remove a task graph and all child tasks.

        Parameters
        __________
        dag: Union[DAG, str]
            Name of the task graph to be dropped or a ``DAG`` instance.
        drop_finalizer: bool
            If true, a finalizer task will also be dropped if it exists.

        Examples
        ________
        Drop a task graph:

        >>> dag_op.drop("your-dag-name")
        """

        def _drop(name: str) -> None:
            task = self.schema.tasks[name]
            with suppress(NotFoundError):
                task.suspend()
                task.drop()

        if drop_finalizer is False:
            # TODO: remove in next major
            warnings.warn(
                "Setting drop_finalizer to False is deprecated. This argument will be removed in the next major "
                "version and finalizer task will always be dropped alongside the DAG.",
                category=DeprecationWarning,
                stacklevel=1,
            )

        dag_name = dag.name if isinstance(dag, DAG) else dag
        root_task = self.schema.tasks[dag_name]
        try:
            root_task_details = root_task.fetch()
        except NotFoundError:
            logger.info("Root task '%s' already dropped", dag_name)
            return

        if root_task_details.state == "started":
            root_task.suspend()

        tasks = self.schema.tasks[dag_name].fetch_task_dependents()
        finalizer = self._extract_finalizer(root_task_details)
        for t in reversed(tasks):
            if finalizer and t.name == finalizer.name and not drop_finalizer:
                # TODO: remove in next major together with deprecation warnings
                continue
            _drop(t.name)

    @api_telemetry
    def deploy(self, dag: DAG, mode: CreateMode = CreateMode.error_if_exists) -> None:
        """Deploys (create) this task graph including all child tasks under a specific schema in Snowflake.

        Parameters
        __________
        dag: The :class:`DAG` instance.
        mode: CreateMode, optional
            One of the following strings.

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the task already exists in Snowflake. Equivalent to SQL ``create task <name> ...``.

            ``CreateMode.or_replace``: Replace if the task already exists in Snowflake. Equivalent to SQL
            ``create or replace task <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the task already exists in Snowflake.
            Equivalent to SQL ``create task <name> if not exists...``

            Default value is ``CreateMode.error_if_exists``.
        """

        def topological_sort() -> Iterator[DAGTask]:
            """Topologically sort by using Kahn algorithm.

            Sort with predecessors then reverse the sorted result.
            """
            from collections import defaultdict, deque

            # Step 1: Compute in-degree for each vertex.
            in_degree: defaultdict[DAGTask, int] = defaultdict(int)
            for node in dag.tasks:
                for neighbor in node.predecessors:
                    in_degree[neighbor] += 1

            # Step 2: Initialize a queue and add nodes with in-degree 0.
            queue: deque[DAGTask] = deque()
            for node in dag.tasks:
                if in_degree[node] == 0:
                    queue.append(node)

            # Step 3: Process nodes in the queue and update in-degrees.
            sorted_nodes: list[DAGTask] = []
            while queue:
                node = queue.popleft()
                sorted_nodes.append(node)
                for neighbor in node.predecessors:
                    in_degree[neighbor] -= 1
                    if in_degree[neighbor] == 0:
                        queue.append(neighbor)

            # Step 4: Check for a cycle in the graph.
            if len(sorted_nodes) != len(dag.tasks):
                raise ValueError("There is a cycle in the task graph.")

            return reversed(sorted_nodes)

        mode = CreateMode[mode]
        if mode is CreateMode.or_replace:
            with suppress(NotFoundError):
                self.drop(dag)

        self.schema.tasks.create(dag._to_low_level_task())
        for dag_task in topological_sort():
            # 2023-10-04(bwarsaw) TODO: pass the enum value to mode instead of the string.
            self.schema.tasks.create(dag_task._to_low_level_task(), mode=mode)
            self.schema.tasks[dag_task.full_name].resume()
        if dag.schedule is not None:
            self.schema.tasks[dag.name].resume()

    @api_telemetry
    def run(self, dag: Union[DAG, str], *, retry_last: bool = False) -> None:
        """Execute the task graph once without waiting for the schedule.

        It resumes the task graph then executes it.

        Parameters
        __________
        retry_last: bool, optional
            Re-execute the last failed task of the task graph. Default is False.

        Examples
        ________
        Run a task graph:

        >>> dag_op.run("your-dag-name")
        """
        schedule = dag.schedule if isinstance(dag, DAG) else self.schema.tasks[dag].fetch().schedule
        if schedule is not None:
            self.schema.tasks[dag.name if isinstance(dag, DAG) else dag].resume()
        self.schema.tasks[dag.name if isinstance(dag, DAG) else dag].execute(retry_last=retry_last)

    @api_telemetry
    def get_complete_dag_runs(self, dag: Union[DAG, str], *, error_only: bool = True) -> list[DAGRun]:
        """Get the complete task graph runs within 60 minutes in Snowflake.

        Parameters
        __________
        dag: Union[DAG, str]
            The task graph to get the task graph runs.
        error_only: bool, optional
            If True, only return the task graph runs that have failed. Default is True.

        Examples
        ________
        Get the complete task graph runs:

        >>> dag_op.get_complete_dag_runs("your-dag-name")
        """
        return [
            DAGRun._from_taskrun(taskrun)
            for taskrun in self.schema.tasks[dag.name if isinstance(dag, DAG) else dag].get_complete_graphs(
                error_only=error_only
            )
        ]

    @api_telemetry
    def get_current_dag_runs(self, dag: Union[DAG, str]) -> list[DAGRun]:
        """Get the current task graph runs or next schedule dag run for next 8 days in Snowflake.

        Parameters
        __________
        dag: Union[DAG, str]
            The task graph to get the task graph runs.

        Examples
        ________
        Get the current task graph runs:

        >>> dag_op.get_current_dag_runs("your-dag-name")
        """
        return [
            DAGRun._from_taskrun(taskrun)
            for taskrun in self.schema.tasks[dag.name if isinstance(dag, DAG) else dag].get_current_graphs()
        ]

    @staticmethod
    def _extract_finalizer(task: Task) -> Optional[FQN]:
        if task.task_relations:
            identifier = json.loads(task.task_relations).get("FinalizerTask")
            return FQN.from_string(identifier) if identifier else None
        return None


__all__ = ["DAG", "DAGTask", "DAGRun", "DAGOperation"]
